"""
================================================
Convolutional Poisson Gamma Belief Network Demo
Chaojie Wang  Sucheng Xiao  Bo Chen  and  Mingyuan Zhou
Published in International Conference on Machine Learning 2019

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>
# License: BSD-3-Clause
import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import CPGBN

import _pickle as cPickle

DATA = cPickle.load(open("data/TREC.pkl", "rb"), encoding='iso-8859-1')

data_vab_list = DATA['Vocabulary']
data_vab_count_list = DATA['Vab_count']
data_vab_length = DATA['Vab_Size']
data_label = DATA['Label']
data_train_list = DATA['Train_Origin']
data_train_label = np.array(DATA['Train_Label'])
data_train_split = DATA['Train_Word_Split']
data_train_list_index = DATA['Train_Word2Index']
data_test_list = DATA['Test_Origin']
data_test_label = np.array(DATA['Test_Label'])
data_test_split = DATA['Test_Word_Split']
data_test_list_index = DATA['Test_Word2Index']
data_value = 25

# ======================= Preprocess =======================#
delete_count = 0

for i in range(len(data_train_list)):

    x_single = np.reshape(data_train_list_index[i], [len(data_train_list_index[i])]).astype(np.int32)
    x_len = x_single.shape[0]

    i_index = i - delete_count
    if i_index == 0:
        batch_len = np.array([x_len])
        batch_rows = x_single
        batch_cols = np.arange(x_len)
        batch_file_index = np.ones_like(x_single) * i_index
        batch_value = np.ones_like(x_single) * data_value
        batch_label = np.array([data_train_label[i]])
    else:
        batch_len = np.concatenate((batch_len, np.array([x_len])), axis=0)
        batch_rows = np.concatenate((batch_rows, x_single), axis=0)
        batch_cols = np.concatenate((batch_cols, np.arange(x_len)), axis=0)
        batch_file_index = np.concatenate((batch_file_index, np.ones_like(x_single) * i_index), axis=0)
        batch_value = np.concatenate((batch_value, np.ones_like(x_single) * data_value), axis=0)
        batch_label = np.concatenate((batch_label, np.array([data_train_label[i]])), axis=0)


batch_len_tr = batch_len
batch_rows_tr = batch_rows
batch_cols_tr = batch_cols
batch_file_index_tr = batch_file_index
batch_value_tr = batch_value
batch_label_tr = batch_label

# ======================= Preprocess =======================#
delete_count = 0

for i in range(len(data_test_list)):

    x_single = np.reshape(data_test_list_index[i], [len(data_test_list_index[i])]).astype(np.int32)
    x_len = x_single.shape[0]

    i_index = i - delete_count
    if i_index == 0:
        batch_len = np.array([x_len])
        batch_rows = x_single
        batch_cols = np.arange(x_len)
        batch_file_index = np.ones_like(x_single) * i_index
        batch_value = np.ones_like(x_single) * data_value
        batch_label = np.array([data_test_label[i]])
    else:
        batch_len = np.concatenate((batch_len, np.array([x_len])), axis=0)
        batch_rows = np.concatenate((batch_rows, x_single), axis=0)
        batch_cols = np.concatenate((batch_cols, np.arange(x_len)), axis=0)
        batch_file_index = np.concatenate((batch_file_index, np.ones_like(x_single) * i_index), axis=0)
        batch_value = np.concatenate((batch_value, np.ones_like(x_single) * data_value), axis=0)
        batch_label = np.concatenate((batch_label, np.array([data_test_label[i]])), axis=0)

batch_len_te = batch_len
batch_rows_te = batch_rows
batch_cols_te = batch_cols
batch_file_index_te = batch_file_index
batch_value_te = batch_value
batch_label_te = batch_label


# data = sio.loadmat('./mnist_gray')
# train_data = np.array(np.ceil(data['train_mnist']*25), order='C')[:, 0:999]
# test_data = np.array(np.ceil(data['train_mnist']*25), order='C')[:, 1000:1999]
# train_data = np.transpose(np.transpose(train_data).reshape([train_data.shape[1],  28, 28]), [0, 2, 1])
# test_data = np.transpose(np.transpose(test_data).reshape([test_data.shape[1], 28, 28]), [0, 2, 1])
# train_label = data['train_label'][:999]
# test_label = data['train_label'][1000:1999]

# create the model and deploy it on gpu or cpu
model = CPGBN([200, 100, 50], 'gpu')
# mode 1, dense input
model.initial([batch_file_index_tr, batch_rows_tr, batch_cols_tr, batch_value_tr], [len(data_train_list) - delete_count, DATA['Vab_Size'], np.max(batch_len_tr)])  # use the shape of train_data to initialize the params of model
train_local_params = model.train(100, [batch_file_index_tr, batch_rows_tr, batch_cols_tr, batch_value_tr], [len(data_train_list) - delete_count, DATA['Vab_Size'], np.max(batch_len_tr)])
train_local_params = model.test(100, [batch_file_index_tr, batch_rows_tr, batch_cols_tr, batch_value_tr], [len(data_train_list) - delete_count, DATA['Vab_Size'], np.max(batch_len_tr)])
test_local_params = model.test(100, [batch_file_index_te, batch_rows_te, batch_cols_te, batch_value_te], [len(data_test_list) - delete_count, DATA['Vab_Size'], np.max(batch_len_te)])

train_theta = np.sum(np.sum(train_local_params.W_nk, axis=3), axis=2).T
test_theta = np.sum(np.sum(test_local_params.W_nk, axis=3), axis=2).T

# Score of test dataset's Theta: 0.682
results = ACC(train_theta, test_theta, batch_label_tr, batch_label_te, 'SVM')
model.save()



import numpy as np
from gensim.test.utils import common_corpus, common_dictionary
from gensim.models.coherencemodel import CoherenceModel

"""
Examples
---------
One way of using this feature is through providing a trained topic model. A dictionary has to be explicitly provided
if the model does not contain a dictionary already

.. sourcecode:: pycon
	#
	# >>> from gensim.test.utils import common_corpus, common_dictionary
	# >>> from gensim.models.ldamodel import LdaModel
	# >>> from gensim.models.coherencemodel import CoherenceModel
	# >>>
	# >>> model = LdaModel(common_corpus, 5, common_dictionary)
	# >>>
	# >>> cm = CoherenceModel(model=model, corpus=common_corpus, coherence='u_mass')
	# >>> coherence = cm.get_coherence()  # get coherence value

Another way of using this feature is through providing tokenized topics such as:

.. sourcecode:: pycon

	# >>> from gensim.test.utils import common_corpus, common_dictionary
	# >>> from gensim.models.coherencemodel import CoherenceModel
	# >>> topics = [
	# ...     ['human', 'computer', 'system', 'interface'],
	# ...     ['graph', 'minors', 'trees', 'eps']
	# ... ]
	# >>>
	# >>> cm = CoherenceModel(topics=topics, corpus=common_corpus, dictionary=common_dictionary, coherence='u_mass')
	# >>> coherence = cm.get_coherence()  # get coherence value

（Please visit https://radimrehurek.com/gensim/models/coherencemodel.html for more usage.）

"""

class Topic_Coherence(object):
	def __init__(self, model=None, topics=None, texts=None, corpus=None, dictionary=None,
					 window_size=None, keyed_vectors=None, coherence='c_v', topn=20, processes=-1):
		'''
		Inputs:
			model : :class:`~gensim.models.basemodel.BaseTopicModel`, optional
				Pre-trained topic model, should be provided if topics is not provided.
				Currently supports :class:`~gensim.models.ldamodel.LdaModel`,
				:class:`~gensim.models.ldamulticore.LdaMulticore`, :class:`~gensim.models.wrappers.ldamallet.LdaMallet` and
				:class:`~gensim.models.wrappers.ldavowpalwabbit.LdaVowpalWabbit`.
				Use `topics` parameter to plug in an as yet unsupported model.
			topics : list of list of str, optional
				List of tokenized topics, if this is preferred over model - dictionary should be provided.
			texts : list of list of str, optional
				Tokenized texts, needed for coherence models that use sliding window based (i.e. coherence=`c_something`)
				probability estimator .
			corpus : iterable of list of (int, number), optional
				Corpus in BoW format.
			dictionary : :class:`~gensim.corpora.dictionary.Dictionary`, optional
				Gensim dictionary mapping of id word to create corpus.
				If `model.id2word` is present, this is not needed. If both are provided, passed `dictionary` will be used.
			window_size : int, optional
				Is the size of the window to be used for coherence measures using boolean sliding window as their
				probability estimator. For 'u_mass' this doesn't matter.
				If None - the default window sizes are used which are: 'c_v' - 110, 'c_uci' - 10, 'c_npmi' - 10.
			coherence : {'u_mass', 'c_v', 'c_uci', 'c_npmi'}, optional
				Coherence measure to be used.
				Fastest method - 'u_mass', 'c_uci' also known as `c_pmi`.
				For 'u_mass' corpus should be provided, if texts is provided, it will be converted to corpus
				using the dictionary. For 'c_v', 'c_uci' and 'c_npmi' `texts` should be provided (`corpus` isn't needed)
			topn : int, optional
				Integer corresponding to the number of top words to be extracted from each topic.
			processes : int, optional
				Number of processes to use for probability estimation phase, any value less than 1 will be interpreted as
				num_cpus - 1.

		Outputs:
			topic_coherence : [float], The topic coherence with model

		'''
		self.model = model
		self.topics = topics
		self.texts = texts
		self.corpus = corpus
		self.dictionary = dictionary
		self.window_size = window_size
		self.keyed_vectors = keyed_vectors
		self.coherence = coherence
		self.topn = topn
		self.processes = processes

		self._get()
		print(f'The topic coherence score is: {self._topic_coherence:.4f}')

	def _get(self):

		cm = CoherenceModel(model=self.model, topics=self.topics, texts=self.texts, corpus=self.corpus,
							dictionary=self.dictionary, window_size=self.window_size, keyed_vectors=self.keyed_vectors,
							coherence=self.coherence, topn=self.topn, processes=self.processes)

		self._topic_coherence = cm.get_coherence()




import numpy as np
from scipy.optimize import linear_sum_assignment

# from sklearn.utils.linear_assignment_ import linear_assignment
# from sklearn.metrics.cluster import normalized_mutual_info_score as NMI, \
#         adjusted_mutual_info_score as AMI, adjusted_rand_score as AR, silhouette_score as SI, calinski_harabasz_score as CH

class Cluster_ACC(object):

	def __init__(self, y, ypred):
		'''
        Inputs:
            y: the ground_true, shape:(n_sample,)
               ypred: pred_label, shape:(n_sample,)

        Outputs:
            accuracy of cluster, in [0, 1]
		'''
		self.y = y
		self.ypred = ypred

		self._get()

		print(f'The cluster accuracy is: {self._cluster_acc:.4f}')

	def _get(self):
		s = np.unique(self.ypred)
		t = np.unique(self.y)

		N = len(np.unique(self.ypred))
		C = np.zeros((N, N), dtype=np.int32)
		for i in range(N):
			for j in range(N):
				idx = np.logical_and(self.ypred == s[i], self.y == t[j])
				C[i][j] = np.count_nonzero(idx)

		# convert the C matrix to the 'true' cost
		Cmax = np.amax(C)
		C = Cmax - C
		indices = linear_sum_assignment(C)
		row = indices[:][:, 0]
		col = indices[:][:, 1]
		# calculating the accuracy according to the optimal assignment
		count = 0
		for i in range(N):
			idx = np.logical_and(self.ypred == s[row[i]], self.y == t[col[i]])
			count += np.count_nonzero(idx)

		self._cluster_acc = 1.0 * count / len(self.y)

		# y_true = y_true.astype(np.int64)
		# assert y_pred.size == y_true.size
		# D = max(y_pred.max(), y_true.max()) + 1
		# w = np.zeros((D, D), dtype=np.int64)
		# for i in range(y_pred.size):
		#     w[y_pred[i], y_true[i]] += 1
		# from sklearn.utils.linear_assignment_ import linear_assignment
		# ind = linear_assignment(w.max() - w) # Optimal label mapping based on the Hungarian algorithm
		# return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size






"""
===========================================
Metric
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu
# License: BSD-3-Clause

import numpy as np

realmin = 2.2e-10

def log_max(x):
    return np.log(np.maximum(x, realmin))

def cosine_simlarity(A, B):
    # A: N*D, B: N*D
    [N, D] = A.shape
    inter_product = np.matmul(A, np.transpose(B))  # N*N
    len_A = np.sqrt(np.sum(A * A, axis=1, keepdims=True))
    len_B = np.sqrt(np.sum(B * B, axis=1, keepdims=True))
    len_AB = np.matmul(len_A, np.transpose(len_B))
    cos_AB = inter_product / (len_AB + realmin)
    cos_AB[(np.arange(N), np.arange(N))] = 1
    return cos_AB




"""
===========================================
Convolutional Poisson Factor Analysis
Chaojie Wang  Sucheng Xiao  Bo Chen  and  Mingyuan Zhou
Published in International Conference on Machine Learning 2019

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import time
import os
import copy
import numpy as np

from ._basic_model import Basic_Model, Params
from .._sampler import Basic_Sampler
from .._utils import *


class CPFA(Basic_Model):

    def __init__(self, K: int, device='gpu'):
        """
        The basic model for CPFA
        Inputs:
            K      : [int] number of kernels in CPFA;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(CPFA, self).__init__()
        setattr(self, '_model_name', 'CPFA')

        self._model_setting.K = K
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)
        # self._sampler.gamma = np.random.gamma

    def initial(self, *args):
        '''
        Inintial the parameters of CPFA with the input documents
        Inputs:
            dense matrix representation
                args[0] : [np.ndarray] N*V*L matrix, N documents represented as V*L sparse matrices

            sparse matrix representation
                args[0] : [list] N documents under the sparse representation, batch.rows, batch.cols, batch.values
                args[1] : [list] the shape of the input document matrix, N*V*L

        Attributes:
            @public:
                global_params.D_k  : [np.ndarray] K*1*S3*S4, K probabilistic convolutional kernels, where the size of each kernel is S3*S4

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
                _model_setting.L        : [int] scalar, the max length of the document
                _hyper_params.D_k_eta   : [float] scalar, parameter in the prior of D_k
                _hyper_params.W_nk_gamma: [float] scalar, parameter in the prior of W_nk
                _hyper_params.p_n_a0    : [float] scalar, parameter in the prior of p_n
                _hyper_params.p_n_b0    : [float] scalar, parameter in the prior of p_n
                _hyper_params.c_n_b0    : [float] scalar, parameter in the prior of c_n
                _hyper_params.c_n_b0    : [float] scalar, parameter in the prior of c_n

        '''
        assert len(args) <= 2, 'Data type error: the input data should be a 3-D np.ndarray or two lists to store the input data under ' \
                              'the sparse representation'
        if len(args) == 1:
            _, self._model_setting.V, self._model_setting.L = args[0].shape
        elif len(args) == 2:
            _, self._model_setting.V, self._model_setting.L = args[1]

        self._model_setting._structure = Params()
        _structure = self._model_setting._structure
        _structure.K_V1 = self._model_setting.V
        _structure.K_V2 = self._model_setting.L + 2  # padding
        _structure.K_S3 = self._model_setting.V
        _structure.K_S4 = 3
        _structure.K_S1 = _structure.K_V1 + 1 - _structure.K_S3
        _structure.K_S2 = _structure.K_V2 + 1 - _structure.K_S4

        self._hyper_params.D_k_eta = 0.05
        self._hyper_params.W_nk_gamma = 0.1 * np.ones([self._model_setting.K, _structure.K_S1, _structure.K_S2])
        self._hyper_params.p_n_a0 = 0.1
        self._hyper_params.p_n_b0 = 0.1
        self._hyper_params.c_n_e0 = 0.1
        self._hyper_params.c_n_f0 = 0.1

        self.global_params.D_k = np.random.rand(self._model_setting.K, 1, _structure.K_S3, _structure.K_S4)
        for k in range(self._model_setting.K):
            self.global_params.D_k[k, :, :, :] = self.global_params.D_k[k, :, :, :] / np.sum(self.global_params.D_k[k, :, :, :])


    def train(self, iter_all: int, *args, **kwargs):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            dense matrix representation
                args[0] : [np.ndarray] N*V*L matrix, N documents represented as V*L sparse matrices

            sparse matrix representation
                args[0] : [list] N documents under the sparse representation, batch.rows, batch.cols, batch.values
                args[1] : [list] the shape of the input document matrix, N*V*L
            **kwargs:
                kwargs['is_train']: [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.W_nk     : [np.ndarray] N_train*K*N_S1*N_S2 matrix, N_train*K feature maps with a size of N_S1*N_S2
                local_params.c_n      : [np.ndarray] N_train vector, the variable in the scale parameter of W_nk
                local_params.p_n      : [np.ndarray] N_train vector, the variable in the scale parameter of W_nk

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of gibbs sampling

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert len(args) <= 2, 'Data type error: the input data should be a 3-D np.ndarray or two lists to store the input data under ' \
                              'the sparse representation'
        if len(args) == 1:
            self._model_setting.N, self._model_setting.V, self._model_setting.L = args[0].shape
            batch_file_indices, batch_rows, batch_cols = np.where(args[0])
            batch_values = args[0][batch_file_indices, batch_rows, batch_cols]
        elif len(args) == 2:
            self._model_setting.N, self._model_setting.V, self._model_setting.L = args[1]
            batch_file_indices, batch_rows, batch_cols, batch_values = args[0]
        if 'is_train' in kwargs:
            is_train = kwargs['is_train']
        else:
            is_train = True
        self._model_setting.Iteration = iter_all

        # initial local parameters
        _structure = self._model_setting._structure
        self.local_params.W_nk = np.random.rand(self._model_setting.N, self._model_setting.K, _structure.K_S1, _structure.K_S2)
        self.local_params.c_n = 1 * np.ones([self._model_setting.N])
        self.local_params.p_n = 1 / (1 + self.local_params.c_n)

        # gibbs sampling
        for iter in range(self._model_setting.Iteration):
            start_time = time.time()

            # data augmentation
            W_nk_aug, D_k_aug = self._sampler.conv_multi_aug(batch_rows, batch_cols, batch_file_indices, batch_values,
                                                              self.global_params.D_k, self.local_params.W_nk)

            # update global_params
            if is_train:
                for k in range(self._model_setting.K):
                    self.global_params.D_k[k, :, :, :] = (D_k_aug[k, :, :, :] + self._hyper_params.D_k_eta) / np.sum(D_k_aug[k, :, :, :] + self._hyper_params.D_k_eta)

            # update local_params
            self.local_params.c_n = self._sampler.gamma(self._hyper_params.c_n_e0 + np.sum(self._hyper_params.W_nk_gamma))
            self.local_params.c_n = self.local_params.c_n / (self._hyper_params.c_n_f0 + np.sum(np.sum(np.sum(W_nk_aug, axis=3), axis=2), axis=1) + realmin)
            self.local_params.p_n = 1 / (self.local_params.c_n + 1)

            self.local_params.W_nk = self._sampler.gamma(W_nk_aug + self._hyper_params.W_nk_gamma) / (1 + self.local_params.c_n)[:, np.newaxis, np.newaxis, np.newaxis]

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, *args):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            dense matrix representation
                args[0] : [np.ndarray] N*V*L matrix, N documents represented as V*L sparse matrices

            sparse matrix representation
                args[0] : [list] N documents under the sparse representation, batch.rows, batch.cols, batch.values
                args[1] : [list] the shape of the input document matrix, N*V*L

        '''
        local_params = self.train(iter_all, *args, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/PFA.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])






import inspect
import numpy as np


def getSamplerFun(sampler, distributions):
    """
    get distributions' function by class sampler.
    Inputs:
        sampler       : sampler class, pydpm._sampler.Basic_Sampler;
        distributions : [str] list of distributions;
    Outputs:
        funcs         : {str: fun}, dict of distributions' functions;
    """
    # get funcs
    funcs = {}
    for dist in distributions:
        funcs[dist] = getattr(sampler, dist)

    return funcs


def getSamplerParams(funcs, distributions, shape):
    """
    get distributions' params by sampler' funcs.
    Inputs:
        funcs         : {str: fun}, dict of distributions' functions;
        distributions : [str] list of distributions;
        shape         : int or tuple, params' shape.
    Outputs:
        params        : {str: [params]}, list of params of each distribution.
    """
    # get params
    int_1 = np.ones(shape, dtype=np.int32, order='C')
    float_05 = np.ones(shape, dtype=np.float32, order='C') / 2
    float_50 = np.ones(shape, dtype=np.float32, order='C') * 5
    int_8 = np.ones(shape, dtype=np.int32, order='C') * 8

    params = {}
    params_special = {'f': [int_8, int_1],
                      'hypergeometric': [int_8, np.ones(shape, dtype=np.int32, order='C')*5, int_8],
                      'multinomial': [5, [0.5, 0.3, 0.2], 100000],
                      'pareto': [float_50, float_05],
                      'triangular': [float_05, float_50, np.ones(shape, dtype=np.float32, order='C')],
                      'uniform': [float_50, float_50],
                      'zipf': [float_50], }
    # 特殊，gamma存在shape的分支，geometric存在采样时间的问题
    init_param = {'size': shape}  # , 'times': 1}
    for p in ['a', 'b', 'prob', 'loc', 'scale', 'shape', 'Lambda', 'p', 'nonc', 'lam']:
        init_param[p] = float_05
    for p in ['count', 'degrees', 'customers', 'n1', 'n2', 'r', 'df', 'dfnum', 'dfden']:
        init_param[p] = int_8
    for dist in distributions:  # 通过param自动填充参数
        param = []
        if dist in params_special.keys():
            params[dist] = params_special[dist]
        else:
            print("%12s" % dist, inspect.signature(funcs[dist]).parameters)
            # print(' '*12, inspect.signature(getattr(np.random, dist)).parameters)
            for p in inspect.signature(funcs[dist]).parameters:
                if p != 'times':
                    param.append(init_param[p])
            params[dist] = param

    return params


def getTorchFunandParams(distributions, shape):
    '''
    get distributions' functions and params from torch.distributions.

    Inputs:
        distributions : [str] list of distributions;
        shape         : int or tuple, params' shape.
    Outputs:
        funcs         : {str: fun}, dict of distributions' functions;
        params        : {str: [params]}, list of params of each distribution.
    '''
    import torch
    funcs = {'beta': torch.distributions.beta.Beta,
                   'exponential': torch.distributions.Exponential,
                   'gamma': torch.distributions.gamma.Gamma,
                   'laplace': torch.distributions.laplace.Laplace,
                   'normal': torch.distributions.Normal,
                   'poisson': torch.distributions.Poisson, }
    # beta(a, b)
    # exp(Lambda)
    # gamma(shape, scale)
    # laplace(loc, scale)
    # normal(loc, scale)
    # poisson(lam)
    float_05 = torch.ones(shape) / 2
    params = {'beta': [float_05, float_05],
              'exponential': [float_05],
              'gamma': [float_05, float_05],
              'laplace': [float_05, float_05],
              'normal': [float_05, float_05],
              'poisson': [float_05], }

    return funcs, params


def getTensorflowFunandParams(distributions, shape):
    '''
    get distributions' functions and params from tensorflow.

    Inputs:
        distributions : [str] list of distributions;
        shape         : int or tuple, params' shape.
    Outputs:
        funcs         : {str: fun}, dict of distributions' functions;
        params        : {str: [params]}, list of params of each distribution.
    '''
    import tensorflow as tf
    import tensorflow_probability as tfp

    funcs = {'beta': tfp.distributions.Beta,
                   'exponential': tfp.distributions.Exponential,
                   'gamma': tfp.distributions.gamma.Gamma,
                   'laplace': tfp.distributions.laplace.Laplace,
                   'normal': tfp.distributions.Normal,
                   'poisson': tfp.distributions.Poisson, }
    float_05 = tf.ones(shape) / 2
    params = {'beta': [float_05, float_05],
              'exponential': [float_05],
              'gamma': [float_05, float_05],
              'laplace': [float_05, float_05],
              'normal': [float_05, float_05],
              'poisson': [float_05], }

    return funcs, params


def repeatSamplerFuncs(funcs, params, repeats, type='pydpm'):
    '''
    sample each distributions funcs repeats times and calculate the average times cost.

    Inputs:
        funcs     : [samplerFuncs]
        params    : [[param] of each dist]
        repeats   : int, repeats times;
        typt      : str, cu, numpy, torch or tensorflow
    Outputs:
        avgt      : [float], average times of each sampler funcs.
    '''
    assert type in ['pydpm', 'numpy', 'torch', 'tensorflow']
    avgt = []
    for dist in distributions:
        start_time = time.time()
        for _ in range(repeats):
            if (type == 'pydpm' or type == 'numpy'):
                funcs[dist](*params[dist])
            else:
                funcs[dist](*params[dist]).sample()
        avgt.append((time.time() - start_time) / repeats)
    return avgt


def plotComparision(distributions, title, types, *args):
    num_types = len(types)
    assert (num_types == len(args)), 'types and avgts do not correspond'
    bar_width = 1 / (num_types + 1)
    index = np.arange(len(distributions))
    colors = ['dodgerblue', 'gray', 'peru']
    for i in range(num_types):
        plt.bar(index + i * bar_width, height=args[i], width=bar_width, color=colors[i], label=types[i])
    plt.legend()
    plt.xticks(index + bar_width/2, distributions, rotation=90)
    plt.ylabel('time(s)')
    # plt.title(title)
    plt.tight_layout()
    plt.show()


# -----------------test the accuracy --------------------
if __name__ == "__main__":
    from pydpm._sampler import Basic_Sampler
    import matplotlib.pyplot as plt
    import time

    sampler = Basic_Sampler('gpu')

    # all distributions: a part of 'distributions = sampler.__dict__.keys()'
    '''distributions = ['beta', 'binomial', 'cauchy', 'chisquare', 'crt', 'dirichlet', 'exponential', 'f', 'gamma', 'geometric', 'gumbel',
    'hypergeometric', 'laplace', 'logistic', 'multinomial', 'negative_binomial', 'noncentral_chisquare', 'noncentral_f',
    'normal', 'pareto', 'poisson', 'power', 'rayleigh', 'standard_cauchy', 'standard_exponential',
    'standard_gamma', 'standard_normal', 'standard_uniform', 't', 'triangular', 'uniform', 'weibull', 'zipf']'''

    shape = (999, 128)  # shape of distributions' params
    repeats = 100

    ## Compare with numpy
    distributions = ['beta', 'chisquare', 'exponential', 'f', 'gamma', 'geometric', 'gumbel', 'laplace', 'logistic',
                     'multinomial', 'negative_binomial', 'normal', 'poisson', 'power', 'zipf']  # 'weibull', 'dirichlet', 'pareto'
    funcs = getSamplerFun(sampler, distributions)
    funcs_np = getSamplerFun(np.random, distributions)
    params = getSamplerParams(funcs, distributions, shape)

    avgt_cu = repeatSamplerFuncs(funcs, params, repeats, 'pydpm')
    avgt_np = repeatSamplerFuncs(funcs_np, params, repeats, 'numpy')
    plotComparision(distributions, 'compare sampler speed with numpy', ['pydpm', 'numpy'], avgt_cu, avgt_np)

    ## compare with PyTorch and Tensorflow
    distributions = ['beta', 'exponential', 'gamma', 'laplace', 'normal', 'poisson']
    funcs_cu = getSamplerFun(sampler, distributions)
    params_cu = getSamplerParams(funcs_cu, distributions, shape)
    funcs_torch, params_torch = getTorchFunandParams(distributions, shape)
    funcs_tf, params_tf = getTensorflowFunandParams(distributions, shape)

    avgt_cu = repeatSamplerFuncs(funcs_cu, params_cu, repeats, 'pydpm')
    avgt_torch = repeatSamplerFuncs(funcs_torch, params_torch, repeats, 'torch')
    avgt_tf = repeatSamplerFuncs(funcs_tf, params_tf, repeats, 'tensorflow')
    # avgt_cu = [0.0008358, 0.0002895999, 0.00047686, 0.00042855, 0.00040248, 0.0002757]
    # avgt_torch = [0.024683303833, 0.00413030385, 0.009334292, 0.0006732845, 0.000860307, 0.0043436401]
    # avgt_tf = [0.008836011, 0.00104925394, 0.00438073, 0.00101441, 0.0008372998, 0.00203398]
    plotComparision(distributions, 'compare sampler speed with torch and tensorflow', ['pydpm', 'torch', 'tensorflow'], avgt_cu, avgt_torch, avgt_tf)





import numpy as np
import seaborn as sns
import scipy.stats as stats
from collections import Counter
import matplotlib.pyplot as plt

from pydpm._sampler import Basic_Sampler

def debug_sampler_and_plot():

    sampler = Basic_Sampler('gpu')

    # gamma
    output = sampler.gamma(np.ones(1000)*4.5, 5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 100, 100), stats.gamma.pdf(np.linspace(0, 100, 100), 4.5, scale=5))
    plt.title('gamma(4.5, 5)')
    plt.show()

    # standard_gamma
    output = sampler.standard_gamma(np.ones(1000)*4.5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 20, 100), stats.gamma.pdf(np.linspace(0, 20, 100), 4.5))
    plt.title('standard_gamma(4.5)')
    plt.show()

    # dirichlet
    output = sampler.dirichlet(np.ones(1000)*4.5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    # x = np.linspace(np.min(output), np.max(output), 100)
    # plt.plot(x, stats.dirichlet.pdf(x, alpha=np.ones(100)*4.5))
    plt.title('dirichlet(4.5)')
    plt.show()

    # beta
    output = sampler.beta(np.ones(1000)*0.5, 0.5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 1, 100), stats.beta.pdf(np.linspace(0, 1, 100), 0.5, 0.5))
    plt.title('beta(0.5, 0.5)')
    plt.show()

    # beta(2, 5)
    output = sampler.beta(np.ones(1000)*2, 5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 1, 100), stats.beta.pdf(np.linspace(0, 1, 100), 2, 5))
    plt.title('beta(2, 5)')
    plt.show()

    # normal
    output = sampler.normal(np.ones(1000)*5, np.ones(1000)*2)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-2, 13, 100), stats.norm.pdf(np.linspace(-2, 13, 100), 5, scale=2))
    plt.title('normal(5, 2)')
    plt.show()

    # standard_normal
    output = sampler.standard_normal(1000)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-3, 3, 100), stats.norm.pdf(np.linspace(-3, 3, 100)))
    plt.title('standard_normal()')
    plt.show()

    # uniform
    output = sampler.uniform(np.ones(1000)*(-2), np.ones(1000)*5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-3, 6, 100), stats.uniform.pdf(np.linspace(-3, 6, 100), -2, 7))
    plt.title('uniform(-2, 5)')
    plt.show()

    # standard_uniform
    output = sampler.standard_uniform(1000)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-0.3, 1.3, 100), stats.uniform.pdf(np.linspace(-0.3, 1.3, 100)))
    plt.title('standard_uniform()')
    plt.show()

    # binomial
    output = sampler.binomial(np.ones(1000)*10, np.ones(1000)*0.5)
    plt.figure()
    plt.hist(output, bins=np.max(output)-np.min(output), density=True, range=(np.min(output)-0.5, np.max(output)-0.5))
    # plt.scatter(np.arange(10), stats.binom._pmf(np.arange(10), 10, 0.5), c='orange', zorder=10)
    plt.title('binomial(10, 0.5)')
    plt.show()

    # negative_binomial
    output = sampler.negative_binomial(np.ones(1000)*10, 0.5)
    plt.figure()
    plt.hist(output, bins=np.max(output)-np.min(output), density=True, range=(np.min(output)-0.5, np.max(output)-0.5))
    plt.scatter(np.arange(30), stats.nbinom._pmf(np.arange(30), 10, 0.5), c='orange', zorder=10)
    plt.title('negative_binomial(10, 0.5)')
    plt.show()

    # multinomial
    output = sampler.multinomial(5, [0.8, 0.2], 1000)
    # output = sampler.multinomial([10]*4, [[0.8, 0.2]]*4, 3)
    plt.figure()
    plt.hist(output[0], bins=15, density=True)
    plt.title('multinomial(5, [0.8, 0.2])')
    plt.show()

    a = np.array([np.array([[i] * 6 for i in range(6)]).reshape(-1), np.array(list(range(6)) * 6)]).T
    output = stats.multinomial(n=5, p=[0.8, 0.2]).pmf(a)
    sns.heatmap(output.reshape(6, 6), annot=True)
    plt.ylabel('number of the 1 kind(p=0.8)')
    plt.xlabel('number of the 2 kind(p=0.2)')
    plt.title('stats.multinomial(n=5, p=[0.8, 0.2])')
    plt.show()

    # poisson
    output = sampler.poisson(np.ones(1000)*10)
    plt.figure()
    plt.hist(output, bins=22, density=True, range=(-0.5, 21.5))
    plt.scatter(np.arange(20), stats.poisson.pmf(np.arange(20), 10), c='orange', zorder=10)
    plt.title('poisson(10)')
    plt.show()

    # cauchy
    output = sampler.cauchy(np.ones(1000)*1, 0.5)
    plt.figure()
    plt.hist(output, bins=20, density=True, range=(-5, 7))
    plt.plot(np.linspace(-5, 7, 100), stats.cauchy.pdf(np.linspace(-5, 7, 100), 1, 0.5))
    plt.title('cauchy(1, 0.5)')
    plt.show()

    # standard_cauchy
    output = sampler.standard_cauchy(1000)
    plt.figure()
    plt.hist(output, bins=20, density=True, range=(-7, 7))
    plt.plot(np.linspace(-7, 7, 100), stats.cauchy.pdf(np.linspace(-7, 7, 100)))
    plt.title('standard_cauchy()')
    plt.show()

    # chisquare
    output = sampler.chisquare(np.ones(1000)*10)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 30, 100), stats.chi2.pdf(np.linspace(0, 30, 100), 10))
    plt.title('chisquare(10)')
    plt.show()

    # noncentral_chisquare
    output = sampler.noncentral_chisquare(np.ones(1000)*10, 5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    # nocentral_chi2 = scale^2 * (chi2 + 2*loc*chi + df*loc^2)
    # E(Z) = nonc + df
    # Var(Z) = 2(df+2nonc)
    plt.title('noncentral_chisquare(df=10, nonc=5)')
    plt.show()

    # exponential
    lam = 0.5
    output = sampler.exponential(np.ones(1000)*lam)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0.01, 4, 100), stats.expon.pdf(np.linspace(0.01, 4, 100), scale=0.5))
    plt.title('exponential(0.5)')
    plt.show()

    # standard_exponential
    output = sampler.standard_exponential(1000)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0.01, 8, 100), stats.expon.pdf(np.linspace(0.01, 8, 100)))
    plt.title('standard_exponential()')
    plt.show()

    # f
    output = sampler.f(np.ones(1000)*10, 10)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 8, 100), stats.f.pdf(np.linspace(0, 8, 100), 10, 10))
    plt.title('f(10, 10)')
    plt.show()

    # noncentral_f
    output = sampler.noncentral_f(np.ones(1000)*10, 10, 5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    # E(F) = (m+nonc)*n / (m*(n-2)), n>2.
    # Var(F) = 2*(n/m)**2 * ((m+nonc)**2 + (m+2*nonc)*(n-2)) / ((n-2)**2 * (n-4))
    plt.title('noncentral_f(dfnum=10, dfden=10, nonc=5)')
    plt.show()

    # geometric
    output = sampler.geometric(np.ones(1000)*0.1)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.scatter(np.arange(50), stats.geom.pmf(np.arange(50), p=0.1), c='orange', zorder=10)
    plt.title('geometric(0.1)')
    plt.show()

    # gumbel
    output = sampler.gumbel(np.ones(1000)*5, np.ones(1000)*2)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 20, 100), stats.gumbel_r.pdf(np.linspace(0, 20, 100)+0.01, 5, scale=2))
    plt.title('gumbel(5, 2)')
    plt.show()
    np.random.gumbel()

    # hypergeometric
    output = sampler.hypergeometric(np.ones(1000)*5, 10, 10)
    plt.figure()
    plt.hist(output, bins=np.max(output)-np.min(output), density=True, range=(np.min(output)+0.5, np.max(output)+0.5))
    plt.scatter(np.arange(10), stats.hypergeom(15, 5, 10).pmf(np.arange(10)), c='orange', zorder=10)  # hypergeom(M, n, N), total, I, tiems
    plt.title('hypergeometric(5, 10, 10)')
    plt.show()

    # laplace
    output = sampler.laplace(np.ones(1000)*5, np.ones(1000)*2)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-10, 20, 100), stats.laplace.pdf(np.linspace(-10, 20, 100), 5, scale=2))
    plt.title('laplace(5, 2)')
    plt.show()

    # logistic
    output = sampler.logistic(np.ones(1000)*5, np.ones(1000)*2)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(-10, 20, 100), stats.logistic.pdf(np.linspace(-10, 20, 100), 5, scale=2))
    plt.title('logistic(5, 2)')
    plt.show()

    # power
    output = sampler.power(np.ones(1000)*0.5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 1.5, 100), stats.powerlaw.pdf(np.linspace(0, 1.5, 100), 0.5))
    plt.title('power(0.5)')
    plt.show()

    # zipf
    output = sampler.zipf(np.ones(1000)*1.1)
    counter = Counter(output)
    filter = np.array([[key, counter[key]] for key in counter.keys() if key < 50])
    plt.figure()
    plt.scatter(filter[:, 0], filter[:, 1] / 1000)
    plt.plot(np.arange(1, 50), stats.zipf(1.1).pmf(np.arange(1, 50)))
    plt.title('zipf(1.1)')
    plt.show()

    # pareto
    output = sampler.pareto(np.ones(1000) * 2, np.ones(1000) * 5)
    plt.figure()
    count, bins, _ = plt.hist(output, bins=50, density=True, range=(np.min(output), 100))
    a, m = 2., 5.  # shape and mode
    fit = a * m ** a / bins ** (a + 1)
    plt.plot(bins, max(count) * fit / max(fit), linewidth=2, color='r')
    plt.title('pareto(2, 5)')
    plt.show()

    # rayleigh
    output = sampler.rayleigh(np.ones(1000)*2.0)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 8, 100), stats.rayleigh(scale=2).pdf(np.linspace(0, 8, 100)))
    plt.title('rayleigh(2)')
    plt.show()

    # t
    output = sampler.t(np.ones(1000)*2.0)
    plt.figure()
    plt.hist(output, bins=20, density=True, range=(-6, 6))
    plt.plot(np.linspace(-6, 6, 100), stats.t(2).pdf(np.linspace(-6, 6, 100)))
    plt.title('t(2)')
    plt.show()

    # triangular
    output = sampler.triangular(np.ones(1000)*0.0, 0.3, 1)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 1, 100), stats.triang.pdf(np.linspace(0, 1, 100), 0.3))
    plt.title('triangular(0, 0.3, 1)')
    plt.show()

    # weibull
    output = sampler.weibull(np.ones(1000)*4.5, 5)
    plt.figure()
    plt.hist(output, bins=20, density=True)
    plt.plot(np.linspace(0, 10, 100), stats.weibull_min.pdf(np.linspace(0, 10, 100), 4.5, scale=5))
    plt.title('weibull(4.5, 5)')
    plt.show()


# -----------------test the accuracy --------------------
if __name__ == "__main__":
    plot_all_distribution_example = True  # plot all distribution samplers' example and compare with its pdf/pmf.

    if plot_all_distribution_example:
        debug_sampler_and_plot()





"""
===========================================
Deep Poisson Factor Analysis Demo
Scalable Deep Poisson Factor Analysis for Topic Modeling
Zhe Gan, Changyou Chen, Ricardo Henao, David Carlson, Lawrence Carin
Publised in International Conference on Machine Learning 2015

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import DPFA

# load data
data = sio.loadmat('./data/mnist_gray.mat')
train_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 0:999]
test_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 1000:1999]
train_label = data['train_label'][:999]
test_label = data['train_label'][1000:1999]

# create the model and deploy it on gpu or cpu
model = DPFA([128, 64, 32], 'gpu')  # topics of 3 layers
model.initial(train_data)  # use the shape of train_data to initialize the params of model
burnin, collection = 100, 80
train_local_params = model.train(burnin, collection, train_data)
train_local_params = model.test(burnin, collection, train_data)
test_local_params = model.test(burnin, collection, test_data)

# evaluate the model with classification accuracy
# the demo accuracy can achieve 0.9099
results = ACC(train_local_params.Theta, test_local_params.Theta, train_label, test_label, 'SVM')

# save the model after training
model.save()
# model.load('./save_models/DPFA.npy')






"""
===========================================
Deep Relational Topic Modeling via Graph Poisson Gamma Belief Network
Chaojie Wang, Hao Zhang, Bo Chen, Dongsheng Wang, Zhengjue Wang
Published in Advances in Neural Information Processing System 2020

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Wei Zhao <13279389260@163.com>; Jiawen Wu <wjw19960807@163.com>
# License: BSD-3-Clause

import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *


class GPGBN(Basic_Model):
    def __init__(self, K: list = [128, 64, 32], device='gpu'):
        """
        The basic model for GPGBN
        Inputs:
            K      : [list] number of topics at diffrent layers in PGBN;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting   : [Params] the model settings of the probabilistic model
                _hyper_params    : [Params] the hyper parameters of the probabilistic model
                _model_setting.T : [int] the network depth

        """
        super(GPGBN, self).__init__()
        setattr(self, '_model_name', 'GPGBN')

        self._model_setting.K = K
        self._model_setting.T = len(K)
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)

    def initial(self, data: np.ndarray):
        '''
        Initial the parameters of PGBN with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [list] T (K_t-1)*(K_t) factor loading matrices at different layers

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
                _hyper_params.Phi_eta   : [int] scalar, the parameter in the prior of Phi
                _hyper_params.Theta_r_k : [int] scalar, the parameter in the prior of Theta
                _hyper_params.p_j_a0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.p_j_b0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.c_j_a0    : [int] scalar, the parameter in the prior of c_j
                _hyper_params.c_j_b0    : [int] scalar, the parameter in the prior of c_j

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]

        self.global_params.Phi = []
        self.global_params.U = []
        self._hyper_params.Phi_eta = []
        self._hyper_params.Theta_r_k = np.ones([self._model_setting.K[self._model_setting.T - 1], 1]) / self._model_setting.K[self._model_setting.T - 1]
        self._hyper_params.eta = np.ones(self._model_setting.T) * 0.01
        self._hyper_params.p_j_a0 = 0.01
        self._hyper_params.p_j_b0 = 0.01
        self._hyper_params.c_j_e0 = 1
        self._hyper_params.c_j_f0 = 1  # TODO 最后四个超参对应的代码块 update c_j 被注释掉了

        for t in range(self._model_setting.T):
            self._hyper_params.Phi_eta.append(self._hyper_params.eta[t])
            if t == 0:
                self.global_params.Phi.append(0.2 + 0.8 * np.random.rand(self._model_setting.V, self._model_setting.K[t]))
            else:
                self.global_params.Phi.append(0.2 + 0.8 * np.random.rand(self._model_setting.K[t-1], self._model_setting.K[t]))
            self.global_params.Phi[t] = self.global_params.Phi[t] / np.maximum(realmin, self.global_params.Phi[t].sum(0))
            self.global_params.U.append(self._sampler.gamma(1 * np.ones([self._model_setting.K[t], 1])) / (1 * np.ones([self._model_setting.K[t], 1])))


    def train(self, iter_all: int, data:np.ndarray, data_A:np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta : [list] T (K_t)*(N) topic proportions at different layers
                local_params.c_j   : [list] T+1 1*N vector, the variables in the scale parameter in the Theta
                local_params.p_j   : [list] T+1 1*N vector, the variables in the scale parameter in the Theta

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of gibbs sampling

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        # todo 1.调整注释  2.train/test  3.展示部分代码  4.cosine_simlar
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.N = data.shape[1]
        self._model_setting.Iteration = iter_all

        # initial local params
        self.local_params.Theta = []
        self.local_params.c_j = []
        for t in range(self._model_setting.T):  # from layer 1 to T
            self.local_params.Theta.append(np.ones([self._model_setting.K[t], self._model_setting.N]) / self._model_setting.K[t])
            self.local_params.c_j.append(np.ones([1, self._model_setting.N]))
        self.local_params.c_j.append(np.ones([1, self._model_setting.N]))
        self.local_params.p_j = self._calculate_pj(self.local_params.c_j, self._model_setting.T)

        Xt_to_t1 = []
        WSZS = []

        self.local_params.Sigma = []
        for t in range(self._model_setting.T):
            Xt_to_t1.append(np.zeros(self.local_params.Theta[t].shape))
            WSZS.append(np.zeros(self.global_params.Phi[t].shape))
            self.local_params.Sigma.append(self._sampler.gamma(1 * np.ones([1, self._model_setting.N])) / (1 * np.ones([1, self._model_setting.N])))

        # gibbs sampling
        LH_list = []
        LH_graph_list = []
        U_all = np.zeros([np.sum(self._model_setting.K), self._model_setting.Iteration])

        for iter in range(self._model_setting.Iteration):
            start_time = time.time()

            # update global params: Phi
            for t in range(self._model_setting.T):  # from layer 1 to T
                if t == 0:
                    Xt_to_t1[t], WSZS[t] = self._sampler.multi_aug(data, self.global_params.Phi[t], self.local_params.Theta[t])
                else:
                    Xt_to_t1[t], WSZS[t] = self._sampler.crt_multi_aug(Xt_to_t1[t-1], self.global_params.Phi[t], self.local_params.Theta[t])

                # Xt1 = self._sampler.crt(Xt_to_t1[t-1].astype('double'), np.dot(self.global_params.Phi[t], self.local_params.Theta[t]))
                # Xt_to_t1[t], WSZS[t] = self._sampler.multi_aug(Xt1.astype('double'), self.global_params.Phi[t], self.local_params.Theta[t])

                if is_train:  # todo 源作不分train，且注释了两行
                    self.global_params.Phi[t] = self._update_Phi(WSZS[t], self._hyper_params.Phi_eta[t])


            # downward pass
            M = np.reshape(data_A, [self._model_setting.N * self._model_setting.N, 1])
            M_rate_k = np.zeros([self._model_setting.N * self._model_setting.N, np.sum(np.array(self._model_setting.K))])
            Theta_inter_k = np.zeros([self._model_setting.N * self._model_setting.N, np.sum(np.array(self._model_setting.K))])

            for t in range(self._model_setting.T):
                theta_t = self.local_params.Theta[t]  # K*N
                u_t = self.global_params.U[t]
                sigma_t = self.local_params.Sigma[t]
                start_index = int(np.sum(self._model_setting.K[:t]))
                for k in range(self._model_setting.K[t]):
                    theta_k = theta_t[k: k+1, :].copy()  # todo matmul 有没有加速？
                    theta_inter_k = np.matmul(np.transpose(theta_k * sigma_t), theta_k * sigma_t)
                    theta_inter_k[(np.arange(self._model_setting.N), np.arange(self._model_setting.N))] = 0
                    Theta_inter_k[:, start_index + k: start_index + k + 1] = np.reshape(theta_inter_k, [-1, 1])
                    M_rate_k[:, start_index + k: start_index + k + 1] = np.reshape(theta_inter_k * u_t[k, 0], [-1, 1])

                    # m_rate_k = np.matmul(np.transpose(theta_k*u_t[k, 0]), theta_k)  # theta_k' * U * theta_k
                    # m_rate_k[(np.arange(self._model_setting.N), np.arange(self._model_setting.N))] = 0  # diagnol is zeros
                    # M_rate_k[:, start_index + k: start_index + k+1] = np.reshape(m_rate_k, [-1, 1])

            M_rate_k = M_rate_k / (np.sum(M_rate_k, 1, keepdims=True) + realmin)  # (N*N)*K
            [M_kn, M_vk] = self._sampler.multi_aug(M.astype('double'), M_rate_k, np.ones([np.sum(np.array(self._model_setting.K)), 1]))
            M_ijk = np.reshape(M_vk, [self._model_setting.N, self._model_setting.N, np.sum(np.array(self._model_setting.K))])

            # update c_j todo 源码注释了
            # if iter > 10:
            #     if self._model_setting.T > 1:
            #         for n in range(self._model_setting.N):
            #             self.local_params.p_j[1][0, n] = np.random.beta(Xt_to_t1[0][:, n].sum(0) + self._hyper_params.p_j_a0,
            #                                           self.local_params.Theta[1][:, n].sum(0) + self._hyper_params.p_j_b0)
            #     else:
            #         for n in range(self._model_setting.N):
            #             self.local_params.p_j[1][0, n] = np.random.beta(Xt_to_t1[0][:, n].sum(0) + self._hyper_params.p_j_a0,
            #                                           self._hyper_params.Theta_r_k.sum(0) + self._hyper_params.p_j_b0)
            #     self.local_params.p_j[1] = np.minimum(np.maximum(self.local_params.p_j[1], realmin), 1 - realmin)  # make sure p_j is no so large or so small
            #     self.local_params.c_j[1][:, :] = (1 - self.local_params.p_j[1]) / self.local_params.p_j[1]
            #
            #     for t in [i for i in range(self._model_setting.T + 1) if i > 1]:  # only T>=2 works  ==> for t = 3:T+1
            #         if t == self._model_setting.T:
            #             for n in range(self._model_setting.N):
            #                 self.local_params.c_j[t][0, n] = np.random.gamma(self._hyper_params.Theta_r_k.sum(0) + self._hyper_params.c_j_e0, 1)\
            #                                                  / (self.local_params.Theta[t - 1][:, n].sum(0) + self._hyper_params.c_j_f0)
            #         else:
            #             for n in range(self._model_setting.N):
            #                 self.local_params.c_j[t][0, n] = np.random.gamma(self.local_params.Theta[t][:, n].sum(0) + self._hyper_params.c_j_e0, 1)\
            #                                                  / (self.local_params.Theta[t - 1][:, n].sum(0) + self._hyper_params.c_j_f0)
            #     p_j_tmp = self._sampler._calculate_pj(self.local_params.c_j, self._model_setting.T)
            #     self.local_params.p_j[2:] = p_j_tmp[2:]

            # update U
            M_kj = np.transpose(np.sum(M_ijk, axis=0))  # K*N
            M_kjt = []

            tmp_start = 0
            tmp_end = 0
            for t in range(self._model_setting.T):

                tmp_end += self._model_setting.K[t]
                start_index = int(np.sum(self._model_setting.K[: t]))
                end_index = int(np.sum(self._model_setting.K[: t+1]))

                M_kjt.append(M_kj[start_index:end_index, :])  # K*N
                theta_inter_kjt = Theta_inter_k[:, start_index:end_index]

                M_k = np.sum(M_kjt[t], axis=1) / 2  # K*1
                theta_inter_k = np.sum(theta_inter_kjt, axis=0) / 2  # K*1
                self.global_params.U[t][:, 0] = self._sampler.gamma(M_k + 1) / (1 + theta_inter_k)

                U_all[tmp_start: tmp_end, iter] = self.global_params.U[t][:, 0]

                tmp_start = tmp_end

            # update Theta
            for t in range(self._model_setting.T - 1, -1, -1):  # for t = T:-1 :1

                if t == self._model_setting.T - 1:
                    shape = np.repeat(self._hyper_params.Theta_r_k, self._model_setting.N, axis=1)
                else:
                    shape = np.dot(self.global_params.Phi[t + 1], self.local_params.Theta[t + 1])

                for j in range(self._model_setting.N):
                    sigma_t = self.local_params.Sigma[t].copy()  # Note this copy here
                    theta_t = self.local_params.Theta[t].copy()  # Note this copy here
                    theta_t[:, j] = 0
                    theta_sigma_t = theta_t * sigma_t  # K*N

                    self.local_params.Theta[t][:, j] = self._sampler.gamma(Xt_to_t1[t][:, j] + M_kjt[t][:, j] + shape[:, j], 1)
                    self.local_params.Theta[t][:, j:j + 1] = self.local_params.Theta[t][:, j:j + 1] / (1 + self.local_params.c_j[t][:, j:j + 1] +
                                                                   sigma_t[:, j:j + 1] * self.global_params.U[t] * np.sum(theta_sigma_t, axis=1, keepdims=True))

                    self.local_params.Sigma[t][:, j:j + 1] = self._sampler.gamma(np.sum(M_kjt[t][:, j:j + 1], keepdims=True) + 1, 1)
                    self.local_params.Sigma[t][:, j:j + 1] = self.local_params.Sigma[t][:, j:j + 1] / (np.sum(self.local_params.Theta[t][:, j:j + 1] * self.global_params.U[t] * np.sum(theta_sigma_t, axis=1, keepdims=True)) + 1)

            end_time = time.time()

            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ', f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

            re = np.dot(self.global_params.Phi[0], self.local_params.Theta[0])
            LH = np.sum(data * log_max(re) - re) #- log_max(gamma(data+1)))
            print("    Data Likelihood " + str(LH / self._model_setting.N))
            LH_list.append(LH / self._model_setting.N)

            re_graph = 0
            for t in range(self._model_setting.T):
                re_graph += np.matmul(np.transpose(self.local_params.Sigma[t] * self.local_params.Theta[t] * self.global_params.U[t]), self.local_params.Sigma[t] * self.local_params.Theta[t])
            re_graph[np.arange(self._model_setting.N), np.arange(self._model_setting.N)] = 0
            LH_graph = np.sum(data_A * log_max(re_graph) - re_graph) #- log_max(gamma(data_A+1)))
            print("    Graph Likelihood " + str(LH_graph / self._model_setting.N))
            LH_graph_list.append(LH_graph / self._model_setting.N)

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/PGBN.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            # Phi;  Theta c_j p_j;  K T V N Interation;  Theta_r_k (p_j_a0*4) Phi_eta eta
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def _calculate_pj(self, c_j: list, T: int):
        '''
        calculate p_j from layer 1 to T+1 according to c_j
        Inputs:
            c_j  : [list] T+1 1*N vector, the variables in the scale parameter in the Theta
            T    : [int] network depth
        Outputs:
            p_j  : [list] T+1 1*N vector, the variables in the scale parameter in the Theta

        '''
        p_j = []
        N = c_j[1].size
        p_j.append((1 - np.exp(-1)) * np.ones([1, N]))  # p_j_1
        p_j.append(1 / (1 + c_j[1]))                    # p_j_2

        for t in [i for i in range(T + 1) if i > 1]:    # p_j_3_T+1; only T>=2 works
            tmp = -np.log(np.maximum(1 - p_j[t - 1], realmin))
            p_j.append(tmp / (tmp + c_j[t]))

        return p_j

    def _update_Phi(self, WSZS_t, Eta_t):
        '''
        update Phi_t at layer t
        Inputs:
            WSZS_t  : [np.ndarray]  (K_t-1)*(K_t) count matrix appearing in the likelihood of Phi_t
            Eta_t   : [np.ndarray]  scalar, the variables in the prior of Phi_t
        Outputs:
            Phi_t   : [np.ndarray]  (K_t-1)*(K_t), topic matrix at layer t

        '''
        Phi_t_shape = WSZS_t + Eta_t
        Phi_t = self._sampler.gamma(Phi_t_shape, 1)
        Phi_t = Phi_t / (Phi_t.sum(0) + realmin)

        return Phi_t




"""
===========================================
Metric to evaluate the performance of the classification
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

from sklearn import svm
import numpy as np

class ACC(object):

    def __init__(self, x_tr: np.ndarray, x_te: np.ndarray, y_tr: np.ndarray, y_te: np.ndarray, model='SVM'):
        '''
        Inputs:
            x_tr : [np.ndarray] K*N_train matrix, N_train latent features of length K
            x_te : [np.ndarray] K*N_test matrix, N_test latent features of length K
            y_tr : [np.ndarray] N_train vector, labels of N_train latent features
            y_te : [np.ndarray] N_test vector, labels of N_test latent features

        Outputs:
            accuracy: [float] scalar, the accuracy score

        '''
        self.x_tr = x_tr
        self.x_te = x_te
        self.y_tr = y_tr
        self.y_te = y_te

        if model == 'SVM':
            self._svm()
        else:
            print("Please input metric model correctly. Options: 'SVM'")

        print(f'The classification accuracy with {model} is: {self._accuracy:.4f}')


    def _svm(self):

        self.model = svm.SVC()
        self.model.fit(self.x_tr.T, self.y_tr)
        print(f'Optimization Finished')
        self._accuracy = self.model.score(self.x_te.T, self.y_te)




"""
===========================================
Poisson Factor Analysis DirBN(Dirichlet belief networks) Demo
Dirichlet belief networks for topic structure learning
He Zhao, Lan Du, Wray Buntine, Mingyuan Zhou
Publihsed in Conference and Workshop on Neural Information Processing Systems 2018

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import DirBN

# load data
data = sio.loadmat('./data/mnist_gray.mat')
train_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 0:499]
test_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 1000:1499]
train_label = data['train_label'][:499]
test_label = data['train_label'][1000:1499]

# create the model and deploy it on gpu or cpu
model = DirBN([100, 100], 'gpu')  # topics of each layers
model.initial(train_data)  # use the shape of train_data to initialize the params of model
train_local_params = model.train(90, train_data)
train_local_params = model.test(90, train_data)
test_local_params = model.test(90, test_data)

# evaluate the model with classification accuracy
# the demo accuracy can achieve 0.78
results = ACC(train_local_params.Theta, test_local_params.Theta, train_label, test_label, 'SVM')

# save the model after training
model.save()
# model.load('./save_models/DirBN.npy')





#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages


setup(
    name='pydpm',
    version='2.1.0',
    description='A python library focuses on constructing deep probabilistic models on GPU.',
    py_modules=['pydpm'],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Chaojie Wang, Wei Zhao, Jiawen Wu',
    author_email='xd_silly@163.com',
    maintainer='BoChenGroup',
    maintainer_email='13279389260@163.com',
    license='Apache License Version 2.0',
    packages=find_packages(),
    # package_data={'pydpm': c_package_data},
    # data_files=c_package_data,
    include_package_data=True,  # include all files
    platforms=["Windows", "Linux"],
    url='https://github.com/BoChenGroup/Pydpm',
    requires=['numpy', 'scipy', 'sklearn', 'pycuda'],
    classifiers=[
        'Environment :: GPU :: NVIDIA CUDA',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries'
    ],
)




# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Claus

import platform
import ctypes
import os
import numpy as np

class Basic_Sampler(object):
    def __init__(self, device='cpu', seed=0, *args, **kwargs):
        """
        The basic sampler model for training all probabilistic models in this package
        Attributes:
            @public:


            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(Basic_Sampler, self).__init__()

        assert device in ['cpu', 'gpu'], 'Device Error, device should be "cpu" or "gpu" '
        self.device = device
        self.seed = seed

        system_type = platform.system()
        assert system_type in ['Windows', 'Linux'], 'System Error, system should be "Windows" or "Linux" '
        self.system_type = system_type

        if self.device == 'cpu':
            self._cpu_sampler_initial()

        elif self.device == 'gpu':
            self._gpu_sampler_initial()

    def _cpu_sampler_initial(self):

        from ._distribution_sampler_cpu import distribution_sampler_cpu
        sampler = distribution_sampler_cpu()
        for distribution_name in dir(sampler):
            if distribution_name[0] != '_':
                setattr(self, distribution_name, getattr(sampler, distribution_name))
            else:
                continue


        from ._model_sampler_cpu import model_sampler_cpu
        sampler = model_sampler_cpu(self.system_type)
        for distribution_name in dir(sampler):
            if distribution_name[0] != '_':
                setattr(self, distribution_name, getattr(sampler, distribution_name))
            else:
                continue



    def _gpu_sampler_initial(self):

        from ._distribution_sampler_gpu import distribution_sampler_gpu
        sampler = distribution_sampler_gpu(self.system_type)
        for distribution_name in dir(sampler):
            if distribution_name[0] != '_':
                setattr(self, distribution_name, getattr(sampler, distribution_name))
            else:
                continue

        from ._model_sampler_gpu import model_sampler_gpu
        sampler = model_sampler_gpu(self.system_type)
        for distribution_name in dir(sampler):
            if distribution_name[0] != '_':
                setattr(self, distribution_name, getattr(sampler, distribution_name))
            else:
                continue






"""
===========================================
Deep Poisson Factor Analysis Demo
Scalable Deep Poisson Factor Analysis for Topic Modeling
Zhe Gan, Changyou Chen, Ricardo Henao, David Carlson, Lawrence Carin
Publised in International Conference on Machine Learning 2015
https://github.com/zhegan27/dpfa_icml2015 news_dpfa_sbn_gibbs.m  pfa_dsbn_gibbs.m
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import os
import copy
import time
import math
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *


class DPFA(Basic_Model):
    def __init__(self, K: list, device='cpu'):
        """
        The basic model for DPFA
        Inputs:
            K      : [list] numbers of topics of 3 layers in DPFA(PFA+DSBN+Gibbs);
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(DPFA, self).__init__()
        setattr(self, '_model_name', 'DPFA')
        
        self._model_setting.K1, self._model_setting.K2, self._model_setting.K3 = K
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)


    def initial(self, data: np.ndarray):
        '''
        Inintial the parameters of DPFA with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]

        self.global_params.Phi = np.random.rand(self._model_setting.V, self._model_setting.K1)
        self.global_params.Phi = self.global_params.Phi / np.sum(self.global_params.Phi, 0)


    def train(self, burnin: int, collection: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            burnin     : [int] the iterations of burnin stage
            collection : [int] the iterations of collection stage
            data       : [np.ndarray] V*N_train matrix, N_train bag-of-words vectors with a vocabulary length of V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta  : [np.ndarray] N_train*K matrix, the topic propotions of N train documents

            @private:
                _model_setting.N          : [int] scalar, the number of the documents in the corpus
                _model_setting.burnin     : [int] scalar, the iterations of burnin stage
                _model_setting.collection : [int] scalar, the iterations of collection stage

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model
        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'

        self._model_setting.N = data.shape[1]
        self._model_setting.burnin = burnin
        self._model_setting.collection = collection

        # initial local parameters
        self.local_params.Theta = 1 / self._model_setting.K1 * np.ones((self._model_setting.K1, self._model_setting.N))

        H1train = np.ones((self._model_setting.K1, self._model_setting.N))

        W1 = 0.1 * np.random.randn(self._model_setting.K1, self._model_setting.K2)
        W2 = 0.1 * np.random.randn(self._model_setting.K2, self._model_setting.K3)
        c1 = 0.1 * np.random.randn(self._model_setting.K1, 1)
        c2 = 0.1 * np.random.randn(self._model_setting.K2, 1)
        Pi = 1 / self._model_setting.K3 * np.ones((self._model_setting.K3, 1))
        H3train = (np.tile(Pi, (1, self._model_setting.N)) > np.random.rand(self._model_setting.K3, self._model_setting.N)).astype(int)
        T = np.dot(W2, H3train)
        prob = 1 / (1 + np.exp(-T))
        H2train = (prob >= np.random.rand(self._model_setting.K2, self._model_setting.N)).astype(int)

        c0 = 1
        gamma0 = 1
        eta = 0.05
        r_k = 50 / self._model_setting.K1 * np.ones((self._model_setting.K1, 1))
        p_i_train = 0.5 * np.ones((1, self._model_setting.N))
        e0 = 1e-2
        f0 = 1e-2

        self.mid = dict()
        self.mid['K'] = []
        self.mid['PhiThetaTrain'] = 0
        self.mid['Count'] = 0

        self.result = dict()
        self.result['loglikeTrain'] = []
        self.result['K'] = []
        self.result['PhiThetaTrain'] = 0
        self.result['Count'] = 0
        self.result['Phi'] = np.zeros((self._model_setting.V, self._model_setting.K1))
        self.result['r_k'] = np.zeros((self._model_setting.K1, 1))
        self.result['W1'] = np.zeros((self._model_setting.K1, self._model_setting.K2))
        self.result['W2'] = np.zeros((self._model_setting.K2, self._model_setting.K3))
        self.result['c1'] = np.zeros((self._model_setting.K1, 1))
        self.result['c2'] = np.zeros((self._model_setting.K2, 1))
        self.result['Pi'] = np.zeros((self._model_setting.K3, 1))
        self.result['ThetaTrain'] = np.zeros((self._model_setting.K1, self._model_setting.N))
        self.result['H1train'] = np.zeros((self._model_setting.K1, self._model_setting.N))
        self.result['H2train'] = np.zeros((self._model_setting.K2, self._model_setting.N))
        self.result['H3train'] = np.zeros((self._model_setting.K3, self._model_setting.N))
        self.result['x_kntrain'] = np.zeros((self._model_setting.K1, self._model_setting.N))

        for iter in range(self._model_setting.burnin + self._model_setting.collection):
            start_time = time.time()

            # 1. Sample x_pnk
            x_kntrain, x_pk = self._sampler.multi_aug(data, self.global_params.Phi, self.local_params.Theta)

            # c1 c2 Pi
            c1 = c1.reshape(-1, 1)
            c2 = c2.reshape(-1, 1)
            Pi = Pi.reshape(-1, 1)

            if (iter < np.ceil(self._model_setting.burnin / 40).astype(int)):
                self.local_params.Theta = self._sampler.gamma(x_kntrain + np.dot(r_k, np.ones((1, self._model_setting.N))) * H1train,
                                             np.dot(np.ones((self._model_setting.K1, 1)), p_i_train))
                if is_train:
                    self.global_params.Phi = self._sampler.gamma(eta + x_pk, 1)
                    self.global_params.Phi = self.global_params.Phi / np.sum(self.global_params.Phi, 0)

            else:
                # 2. Sample H1
                lix = (x_kntrain == 0)
                cix, rix = np.where(x_kntrain.T == 0)
                T = np.dot(W1, H2train) + c1
                prob = 1 / (1 + np.exp(-T))
                p1 = prob[lix] * ((1 - p_i_train[0, cix].T) ** r_k[rix].flatten())
                p0 = 1 - prob[lix]
                H1train = np.ones((self._model_setting.K1, self._model_setting.N))
                H1train[lix] = ((p1 / (p1 + p0)) > np.random.rand(len(rix)))

                # 3. inference of sbn
                # (1). update gamma0
                Xmat = np.dot(W1, H2train) + c1
                Xvec = Xmat.reshape(self._model_setting.K1 * self._model_setting.N, 1)
                gamma0vec = self._polya_gam_rnd_truncated(np.ones((self._model_setting.K1 * self._model_setting.N, 1)), Xvec, 20)
                gamma0Train = gamma0vec.reshape(self._model_setting.K1, self._model_setting.N)

                # (2). update W1
                for j in range(self._model_setting.K1):
                    Hgam = H2train * gamma0Train[j, :]
                    invSigmaW = np.eye(self._model_setting.K2) + np.dot(Hgam, H2train.T)
                    MuW = np.dot(np.linalg.inv(invSigmaW), (np.sum(H2train * (H1train[j, :] - 0.5 - c1[j] * gamma0Train[j, :]), 1)))
                    R = self._choll(invSigmaW)
                    W1[j, :] = MuW + np.dot(np.linalg.inv(R), np.random.rand(self._model_setting.K2, 1)).flatten()

                # (3). update H2
                res = np.dot(W1, H2train)
                for k in range(self._model_setting.K2):
                    res = res - np.dot(W1[:, k].reshape(-1, 1), H2train[k, :].reshape(1, -1))
                    mat1 = res + c1
                    vec1 = sum((H1train - 0.5 - gamma0Train * mat1) * W1[:, k].reshape(-1, 1))
                    vec2 = sum(gamma0Train * np.power(W1[:, k].reshape(-1, 1), 2)) / 2
                    logz = vec1 - vec2 + np.dot(W2[k, :], H3train) + c2[k]
                    probz = 1 / (1 + np.exp(-logz))
                    H2train[k, :] = (probz > np.random.rand(1, self._model_setting.N))
                    res = res + np.dot(W1[:, k].reshape(-1, 1), H2train[k, :].reshape(1, -1))

                # (4). update c1
                sigmaC = 1 / (np.sum(gamma0Train, 1) + 1)
                muC = sigmaC * np.sum(H1train - 0.5 - gamma0Train * np.dot(W1, H2train), 1)
                c1 = np.random.normal(muC, np.sqrt(sigmaC))

                # (5). update gamma1
                Xmat = np.dot(W2, H3train) + c2
                Xvec = Xmat.reshape(self._model_setting.K2 * self._model_setting.N, 1)
                gamma1vec = self._polya_gam_rnd_truncated(np.ones((self._model_setting.K2 * self._model_setting.N, 1)), Xvec, 20)
                gamma1Train = gamma1vec.reshape(self._model_setting.K2, self._model_setting.N)

                # (6). update W2
                for k in range(self._model_setting.K2):
                    Hgam = H3train * gamma1Train[k, :]
                    invSigmaW = np.eye(self._model_setting.K3) + np.dot(Hgam, H3train.T)
                    MuW = np.dot(np.linalg.inv(invSigmaW), np.sum(H3train * (H2train[k, :] - 0.5 - c2[k] * gamma1Train[k, :]), 1))
                    R = self._choll(invSigmaW)
                    W2[k, :] = MuW + np.dot(np.linalg.inv(R), np.random.randn(self._model_setting.K3, 1)).flatten()

                # (7). update H3
                res = np.dot(W2, H3train)
                for k in range(self._model_setting.K3):
                    res = res - np.dot(W2[:, k].reshape(-1, 1), H3train[k, :].reshape(1, -1))
                    mat1 = res + c2
                    vec1 = sum((H2train - 0.5 - gamma1Train * mat1) * W2[:, k].reshape(-1, 1))
                    vec2 = sum(gamma1Train * np.power(W2[:, k].reshape(-1, 1), 2)) / 2
                    logz = vec1 - vec2
                    probz = np.exp(logz) * Pi[k] / (np.exp(logz) * Pi[k] + 1 - Pi[k])
                    H3train[k, :] = (probz > np.random.rand(1, self._model_setting.N))
                    res = res + np.dot(W2[:, k].reshape(-1, 1), H3train[k, :].reshape(1, -1))

                Znz = np.sum(H3train, 1)
                Pi = np.random.beta(c0 / self._model_setting.K3 + Znz, c0 - c0 / self._model_setting.K3 + self._model_setting.N - Znz)

                # (8). update c2
                sigmaC = 1 / (np.sum(gamma1Train, 1) + 1)
                muC = sigmaC * np.sum(H2train - 0.5 - gamma1Train * np.dot(W2, H3train), 1)
                c2 = np.random.normal(muC, np.sqrt(sigmaC))

                # 4. Sample r_k
                _, kk = np.where(x_kntrain.T)
                temp = x_kntrain.T.flatten()
                counts = temp[temp != 0]
                ll = np.zeros(len(counts))
                L_k = np.zeros((self._model_setting.K1, 1))
                for k in range(self._model_setting.K1):
                    L_k[k], ll[kk == k] = self._crt(counts[kk == k], r_k[k])
                sumbpi = np.sum(H1train * np.log(1-p_i_train+realmin*((1-p_i_train) < realmin)), 1)
                p_prime_k = -sumbpi / (c0 - sumbpi)
                gamma0 = self._sampler.gamma(e0 + self._crt(L_k.flatten(), np.array([gamma0]))[0], 1 / (f0 - sum(np.log(1 - p_prime_k + realmin * ((1 - p_prime_k) < realmin)))))
                r_k = self._sampler.gamma(gamma0 + L_k.flatten(), 1 / (-sumbpi + c0))

                # 5. Sample Theta
                self.local_params.Theta = self._sampler.gamma(x_kntrain + np.dot(r_k.reshape(-1, 1), np.ones((1, self._model_setting.N))) * H1train,
                                             np.dot(np.ones((self._model_setting.K1, 1)), p_i_train))

                # 6. Sample Phi
                if is_train:
                    self.global_params.Phi = self._sampler.gamma(eta + x_pk, 1)
                    self.global_params.Phi = self.global_params.Phi / np.sum(self.global_params.Phi, 0)

            # Now, collect self.results
            if iter < self._model_setting.burnin:
                X1train = np.dot(self.global_params.Phi, self.local_params.Theta)
                self.mid['PhiThetaTrain'] = self.mid['PhiThetaTrain'] + X1train
                self.mid['Count'] = self.mid['Count'] + 1
                temp = np.count_nonzero(np.sum(x_kntrain, 1))
                self.mid['K'].append(temp)

            else:
                X1train = np.dot(self.global_params.Phi, self.local_params.Theta)
                self.result['PhiThetaTrain'] = self.result['PhiThetaTrain'] + X1train
                self.result['Count'] = self.result['Count'] + 1
                temp = np.count_nonzero(np.sum(x_kntrain, 1))
                self.result['K'].append(temp)

                if is_train:
                    self.result['Phi'] = self.result['Phi'] + self.global_params.Phi / self._model_setting.collection

                self.result['r_k'] = self.result['r_k'] + r_k / self._model_setting.collection
                self.result['W1'] = self.result['W1'] + W1 / self._model_setting.collection
                self.result['W2'] = self.result['W2'] + W2 / self._model_setting.collection
                self.result['c1'] = self.result['c1'] + c1 / self._model_setting.collection
                self.result['c2'] = self.result['c2'] + c2 / self._model_setting.collection
                self.result['Pi'] = self.result['Pi'] + Pi / self._model_setting.collection

                self.result['ThetaTrain'] = self.result['ThetaTrain'] + self.local_params.Theta / self._model_setting.collection
                self.result['H1train'] = self.result['H1train'] + H1train / self._model_setting.collection
                self.result['H2train'] = self.result['H2train'] + H2train / self._model_setting.collection
                self.result['H3train'] = self.result['H3train'] + H3train / self._model_setting.collection
                self.result['x_kntrain'] = self.result['x_kntrain'] + x_kntrain / self._model_setting.collection

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

        return copy.deepcopy(self.local_params)


    def test(self, burnin: int, collection: int, data: np.ndarray):
        '''
        Inputs:
            burnin     : [int] the iterations of burnin stage
            collection : [int] the iterations of collection stage
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(burnin, collection, data, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/DPFA.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])


    def _polya_gam_rnd_truncated(self, a, c, KK, IsBiased=None):
        '''
        Generating Polya-Gamma random varaibles using approximation method
        '''
        IsBiased = False
        x = 1 / 2 / math.pi**2 * np.sum(self._sampler.gamma(np.dot(a, np.ones((1, KK))), 1) /
                                         (np.power((np.array([i for i in range(KK)]) + 0.5), 2) + np.power(c, 2) / 4 / math.pi**2), 1)
        if ~IsBiased:
            temp = abs(c / 2)
            temp[temp<=0] = realmin
            xmeanfull = (np.tanh(temp) / (temp) / 4)
            xmeantruncate = 1 / 2 / math.pi**2 * np.sum(1 / (np.power((np.array([i for i in range(KK)]) + 0.5), 2) + np.power(c, 2)/4/math.pi**2), 1)
            x = x * xmeanfull.flatten() / (xmeantruncate)

        return x


    def _choll(self, A):
        P = A.copy()
        q = np.linalg.cholesky(P)
        q = q.T

        return q


    def _crt(self, x, r):
        xx = np.unique(x)
        jj = np.array([np.argwhere(xx == t) for t in x.flatten()]).flatten()
        L = np.zeros(len(x))
        Lsum = 0
        if not x is None:
            for i in range(len(xx)):
                y = int(xx[i])
                if y > 0:
                    L[jj == i] = np.sum(np.random.rand(np.count_nonzero(jj == i), y) <= (r / (r + [t for t in range(y)])), 1)
            Lsum = int(sum(L))

        return Lsum, L








"""
===========================================
Poisson Factor Analysis
Beta-Negative Binomial Process and Poisson Factor Analysis
Mingyuan Zhou, Lauren Hannah, David Dunson, Lawrence Carin
Publihsed in International Conference on Artificial Intelligence and Statistic 2012

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *

class PFA(Basic_Model):
    def __init__(self, K: int, device='cpu'):
        """
        The basic model for PFA
        Inputs:
            K      : [int] number of topics in PFA;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(PFA, self).__init__()
        setattr(self, '_model_name', 'PFA')

        self._model_setting.K = K
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)


    def initial(self, data: np.ndarray):
        '''
        Inintial the parameters of PFA with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
                _hyper_params.Phi_eta   : [int] scalar, the parameter in the prior of Phi
                _hyper_params.Theta_r_k : [int] scalar, the parameter in the prior of Theta
                _hyper_params.p_j_a0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.p_j_b0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.c_j_a0    : [int] scalar, the parameter in the prior of c_j
                _hyper_params.c_j_b0    : [int] scalar, the parameter in the prior of c_j

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]

        self.global_params.Phi = 0.2 + 0.8 * np.random.rand(self._model_setting.V, self._model_setting.K)
        self.global_params.Phi = self.global_params.Phi / np.maximum(realmin, np.sum(self.global_params.Phi, axis=0))

        self._hyper_params.Phi_eta = 0.01
        self._hyper_params.Theta_r_k = np.ones([self._model_setting.K, 1]) / self._model_setting.K
        self._hyper_params.p_j_a0 = 0.01
        self._hyper_params.p_j_b0 = 0.01
        self._hyper_params.c_j_a0 = 1
        self._hyper_params.c_j_b0 = 1


    def train(self, iter_all: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta    : [np.ndarray] N_train*K matrix, the topic propotions of N_train documents
                local_params.c_j      : [np.ndarray] scalar,
                local_params.p_j      : [np.ndarray] scalar,

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of gibbs sampling

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.N = data.shape[1]
        self._model_setting.Iteration = iter_all

        # initial local parameters
        self.local_params.Theta = np.ones([self._model_setting.K, self._model_setting.N]) / self._model_setting.K
        self.local_params.c_j = np.ones([1, self._model_setting.N])
        self.local_params.p_j = np.ones([1, self._model_setting.N])

        # gibbs sampling
        for iter in range(self._model_setting.Iteration):
            start_time = time.time()

            # update Phi
            Xt_to_t1, WSZS = self._sampler.multi_aug(data, self.global_params.Phi, self.local_params.Theta)

            if is_train:
                self.global_params.Phi = self._update_Phi(WSZS, self._hyper_params.Phi_eta)

            # Update c_j, p_j
            self.local_params.p_j[0, :] = self._sampler.beta(Xt_to_t1.sum(0) + self._hyper_params.p_j_a0, self._hyper_params.Theta_r_k.sum(0) + self._hyper_params.p_j_b0)
            self.local_params.p_j = np.minimum(np.maximum(self.local_params.p_j, realmin), 1 - realmin)  # make sure p_j is not too large or small
            self.local_params.c_j = (1 - self.local_params.p_j) / self.local_params.p_j

            # Update Theta
            shape = np.repeat(self._hyper_params.Theta_r_k, self._model_setting.N, axis=1)
            self.local_params.Theta = self._update_Theta(Xt_to_t1, shape, self.local_params.c_j, self.local_params.p_j)

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/PFA.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])


    def _update_Phi(self, WSZS_t, Eta_t):
        '''
        update Phi_t at layer t
        Inputs:
            WSZS_t  : [np.ndarray]  (K_t-1)*(K_t) count matrix appearing in the likelihood of Phi_t
            Eta_t   : [np.ndarray]  scalar, the variables in the prior of Phi_t
        Outputs:
            Phi_t   : [np.ndarray]  (K_t-1)*(K_t), topic matrix at layer t

        '''
        Phi_t_shape = WSZS_t + Eta_t
        Phi_t = self._sampler.gamma(Phi_t_shape, 1)
        Phi_t = Phi_t / (Phi_t.sum(0) + realmin)

        return Phi_t


    def _update_Theta(self, Xt_to_t1_t, shape, c_j_t1, p_j_t):
        '''
        update Theta_t at layer t
        Inputs:
            Xt_to_t1_t : [np.ndarray]  (K_t-1)*(K_t) count matrix appearing in the likelihood of Phi_t
            shape      : [np.ndarray]  scalar, the variables in the prior of Phi_t
            c_j_t1     : [np.ndarray]  N * 1 vector, the variables in the scale parameter in the Theta_t+1
            p_j_t      : [np.ndarray]  N * 1 vector, the variables in the scale parameter in the Theta_t
        Outputs:
            Theta_t   : [np.ndarray]  (K_t-1)*(K_t), topic proportion matrix at layer t

        '''
        Theta_t_shape = Xt_to_t1_t + shape
        Theta_t = self._sampler.gamma(Theta_t_shape, 1) / (c_j_t1[0, :] - np.log(np.maximum(realmin, 1 - p_j_t[0, :])))

        return Theta_t




"""
===========================================
Poisson Gamma Belief Network
Mingyuan Zhou, Yulai Cong and Bo Chen
Published in Advances in Neural Information Processing System 2015

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *


class PGBN(Basic_Model):
    def __init__(self, K: list, device='gpu'):
        """
        The basic model for PGBN
        Inputs:
            K      : [list] number of topics at diffrent layers in PGBN;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting   : [Params] the model settings of the probabilistic model
                _hyper_params    : [Params] the hyper parameters of the probabilistic model
                _model_setting.T : [int] the network depth

        """
        super(PGBN, self).__init__()
        setattr(self, '_model_name', 'PGBN')

        self._model_setting.K = K
        self._model_setting.T = len(K)
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)

    def initial(self, data: np.ndarray):
        '''
        Initial the parameters of PGBN with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [list] T (K_t-1)*(K_t) factor loading matrices at different layers

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
                _hyper_params.Phi_eta   : [int] scalar, the parameter in the prior of Phi
                _hyper_params.Theta_r_k : [int] scalar, the parameter in the prior of Theta
                _hyper_params.p_j_a0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.p_j_b0    : [int] scalar, the parameter in the prior of p_j
                _hyper_params.c_j_a0    : [int] scalar, the parameter in the prior of c_j
                _hyper_params.c_j_b0    : [int] scalar, the parameter in the prior of c_j

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]

        self.global_params.Phi = []
        self._hyper_params.Phi_eta = []
        self._hyper_params.Theta_r_k = np.ones([self._model_setting.K[self._model_setting.T - 1], 1]) / self._model_setting.K[self._model_setting.T - 1]
        self._hyper_params.p_j_a0 = 0.01
        self._hyper_params.p_j_b0 = 0.01
        self._hyper_params.c_j_e0 = 1
        self._hyper_params.c_j_f0 = 1

        for t in range(self._model_setting.T):
            self._hyper_params.Phi_eta.append(0.01)
            if t == 0:
                self.global_params.Phi.append(0.2 + 0.8 * np.random.rand(self._model_setting.V, self._model_setting.K[t]))
            else:
                self.global_params.Phi.append(0.2 + 0.8 * np.random.rand(self._model_setting.K[t-1], self._model_setting.K[t]))
            self.global_params.Phi[t] = self.global_params.Phi[t] / np.maximum(realmin, self.global_params.Phi[t].sum(0))


    def train(self, iter_all: int, data:np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta : [list] T (K_t)*(N) topic proportions at different layers
                local_params.c_j   : [list] T+1 1*N vector, the variables in the scale parameter in the Theta
                local_params.p_j   : [list] T+1 1*N vector, the variables in the scale parameter in the Theta

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of gibbs sampling

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.N = data.shape[1]
        self._model_setting.Iteration = iter_all

        # initial local params
        self.local_params.Theta = []
        self.local_params.c_j = []
        for t in range(self._model_setting.T):  # from layer 1 to T
            self.local_params.Theta.append(np.ones([self._model_setting.K[t], self._model_setting.N]) / self._model_setting.K[t])
            self.local_params.c_j.append(np.ones([1, self._model_setting.N]))
        self.local_params.c_j.append(np.ones([1, self._model_setting.N]))
        self.local_params.p_j = self._calculate_pj(self.local_params.c_j, self._model_setting.T)

        Xt_to_t1 = []
        WSZS = []
        for t in range(self._model_setting.T):
            Xt_to_t1.append(np.zeros(self.local_params.Theta[t].shape))
            WSZS.append(np.zeros(self.global_params.Phi[t].shape))

        # gibbs sampling
        for iter in range(self._model_setting.Iteration):
            start_time = time.time()

            # update global params
            for t in range(self._model_setting.T):  # from layer 1 to T
                if t == 0:
                    Xt_to_t1[t], WSZS[t] = self._sampler.multi_aug(data, self.global_params.Phi[t], self.local_params.Theta[t])
                else:
                    Xt_to_t1[t], WSZS[t] = self._sampler.crt_multi_aug(Xt_to_t1[t-1], self.global_params.Phi[t], self.local_params.Theta[t])
                if is_train:
                    self.global_params.Phi[t] = self._update_Phi(WSZS[t], self._hyper_params.Phi_eta[t])

            # update local params
            if self._model_setting.T > 1:
                self.local_params.p_j[1][0, :] = self._sampler.beta(Xt_to_t1[0].sum(0) + self._hyper_params.p_j_a0, self.local_params.Theta[1].sum(0) + self._hyper_params.p_j_b0)
            else:
                self.local_params.p_j[1][0, :] = self._sampler.beta(Xt_to_t1[0].sum(0) + self._hyper_params.p_j_a0, self._hyper_params.Theta_r_k.sum(0) + self._hyper_params.p_j_b0)
            self.local_params.p_j[1] = np.minimum(np.maximum(self.local_params.p_j[1], realmin), 1 - realmin)  # make sure p_j is not too large or small
            self.local_params.c_j[1] = (1 - self.local_params.p_j[1]) / self.local_params.p_j[1]

            for t in [i for i in range(self._model_setting.T + 1) if i > 1]:  # from layer 3 to layer T+1
                if t == self._model_setting.T:
                    self.local_params.c_j[t][0, :] = self._sampler.gamma(self._hyper_params.Theta_r_k.sum(0) + self._hyper_params.c_j_e0, 1) \
                                                      / (self.local_params.Theta[t-1].sum(0) + self._hyper_params.c_j_f0 + realmin)
                else:
                    self.local_params.c_j[t][0, :] = self._sampler.gamma(self.local_params.Theta[t].sum(0) + self._hyper_params.c_j_e0, 1) \
                                                      / (self.local_params.Theta[t-1].sum(0) + self._hyper_params.c_j_f0 + realmin)

            p_j_tmp = self._calculate_pj(self.local_params.c_j, self._model_setting.T)
            self.local_params.p_j[2:] = p_j_tmp[2:]

            for t in range(self._model_setting.T - 1, -1, -1):  # from layer T to 1
                if t == self._model_setting.T - 1:
                    shape = np.repeat(self._hyper_params.Theta_r_k, self._model_setting.N, axis=1)
                else:
                    shape = np.dot(self.global_params.Phi[t + 1], self.local_params.Theta[t + 1])
                self.local_params.Theta[t] = self._update_Theta(Xt_to_t1[t], shape, self.local_params.c_j[t + 1], self.local_params.p_j[t])

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/PGBN.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def _calculate_pj(self, c_j: list, T: int):
        '''
        calculate p_j from layer 1 to T+1 according to c_j
        Inputs:
            c_j  : [list] T+1 1*N vector, the variables in the scale parameter in the Theta
            T    : [int] network depth
        Outputs:
            p_j  : [list] T+1 1*N vector, the variables in the scale parameter in the Theta

        '''
        p_j = []
        N = c_j[1].size
        p_j.append((1 - np.exp(-1)) * np.ones([1, N]))  # p_j_1
        p_j.append(1 / (1 + c_j[1]))                    # p_j_2

        for t in [i for i in range(T + 1) if i > 1]:    # p_j_3_T+1; only T>=2 works
            tmp = -np.log(np.maximum(1 - p_j[t - 1], realmin))
            p_j.append(tmp / (tmp + c_j[t]))

        return p_j


    def _update_Phi(self, WSZS_t, Eta_t):
        '''
        update Phi_t at layer t
        Inputs:
            WSZS_t  : [np.ndarray]  (K_t-1)*(K_t) count matrix appearing in the likelihood of Phi_t
            Eta_t   : [np.ndarray]  scalar, the variables in the prior of Phi_t
        Outputs:
            Phi_t   : [np.ndarray]  (K_t-1)*(K_t), topic matrix at layer t

        '''
        Phi_t_shape = WSZS_t + Eta_t
        Phi_t = self._sampler.gamma(Phi_t_shape, 1)
        Phi_t = Phi_t / Phi_t.sum(0)

        return Phi_t


    def _update_Theta(self, Xt_to_t1_t, shape, c_j_t1, p_j_t):
        '''
        update Theta_t at layer t
        Inputs:
            Xt_to_t1_t : [np.ndarray]  (K_t-1)*(K_t) count matrix appearing in the likelihood of Phi_t
            shape      : [np.ndarray]  scalar, the variables in the prior of Phi_t
            c_j_t1     : [np.ndarray]  N * 1 vector, the variables in the scale parameter in the Theta_t+1
            p_j_t      : [np.ndarray]  N * 1 vector, the variables in the scale parameter in the Theta_t
        Outputs:
            Theta_t   : [np.ndarray]  (K_t-1)*(K_t), topic proportion matrix at layer t

        '''
        Theta_t_shape = Xt_to_t1_t + shape
        Theta_t = self._sampler.gamma(Theta_t_shape, 1) / (c_j_t1[0, :] - np.log(np.maximum(realmin, 1 - p_j_t[0, :])))

        return Theta_t




"""
===========================================
Deep Poisson Gamma Dynamical Systems
Dandan Guo, Bo Chen and Hao Zhang
Published in Neural Information Processing Systems 2018

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause


import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *

class DPGDS(Basic_Model):
    def __init__(self, K: int, device='gpu'):
        """
        The basic model for DPGDS
        Inputs:
            K      : [int] number of topics in DPGDS;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(DPGDS, self).__init__()
        setattr(self, '_model_name', 'DPGDS')

        self._model_setting.K = K
        self._model_setting.L = len(K)
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)

        # self.Crt_Matrix = Model_Sampler_CPU.Crt_Matrix
        # self.Multrnd_Matrix_CPU = Model_Sampler_CPU.Multrnd_Matrix
        # self.Sample_Pi = Model_Sampler_CPU.Sample_Pi
        # self.Multrnd_Matrix = self._sampler.multi_aug


    def initial(self, data: np.ndarray):
        '''
        Inintial the parameters of DPGDS with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V         : [int] scalar, the length of the vocabulary
                _model_setting.Stationary: [int] scalar,
                _hyper_params.Supara     : [dict] scalar,
                _hyper_params.Para       : [dict] scalar,

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]
        self._model_setting.Stationary = 1

        self.global_params.Phi = [0] * self._model_setting.L
        self.global_params.Pi = [0] * self._model_setting.L
        self.global_params.Xi = [0] * self._model_setting.L
        self.global_params.V = [0] * self._model_setting.L
        self.global_params.beta = [0] * self._model_setting.L
        self.global_params.q = [0] * self._model_setting.L
        self.global_params.h = [0] * self._model_setting.L
        self.global_params.n = [0] * self._model_setting.L
        self.global_params.rou = [0] * self._model_setting.L

        for l in range(self._model_setting.L):
            if l == 0:
                self.global_params.Phi[l] = np.random.rand(self._model_setting.V, self._model_setting.K[l])
            else:
                self.global_params.Phi[l] = np.random.rand(self._model_setting.K[l - 1], self._model_setting.K[l])
            self.global_params.Phi[l] = self.global_params.Phi[l] / np.sum(self.global_params.Phi[l], axis=0)

            self.global_params.Pi [l] = np.eye(self._model_setting.K[l])
            self.global_params.Xi[l] = 1
            self.global_params.V[l] = np.ones((self._model_setting.K[l], 1))
            self.global_params.beta[l] = 1
            self.global_params.h[l] = np.zeros((self._model_setting.K[l], self._model_setting.K[l]))
            self.global_params.n[l] = np.zeros((self._model_setting.K[l], 1))
            self.global_params.rou[l] = np.zeros((self._model_setting.K[l], 1))

        self._hyper_params.tao0 = 1
        self._hyper_params.gamma0 = 100
        self._hyper_params.eta0 = 0.1
        self._hyper_params.epilson0 = 0.1

    def train(self, iter_all: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of sampling
            train_data : [np.ndarray] V*N_train matrix, N_train bag-of-words vectors with a vocabulary length of V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta : [np.ndarray] N_train*K matrix, the topic propotions of N_train documents

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Burnin    : [int] scalar, the burunin iterations of sampling
                _model_setting.Collection: [int] scalar, the Collection iterations of sampling
                _hyper_params.Para       : [dict] scalar,
        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.T = data.shape[1]

        # local params
        self.local_params.Theta = [0] * self._model_setting.L
        self.local_params.delta = [0] * self._model_setting.L
        self.local_params.Zeta = [0] * self._model_setting.L

        for l in range(self._model_setting.L):
            self.local_params.Theta[l] = np.ones((self._model_setting.K[l], self._model_setting.T)) / self._model_setting.K[l]
            self.local_params.Zeta[l] = np.zeros((self._model_setting.T + 1, 1))
            self.local_params.delta[l] = np.ones((self._model_setting.T, 1))

        # temporary params
        A_KT = [0] * self._model_setting.L
        A_VK = [0] * self._model_setting.L
        L_dotkt = [0] * self._model_setting.L
        L_kdott = [0] * self._model_setting.L
        L_KK = [0] * self._model_setting.L
        prob1 = [0] * self._model_setting.L
        prob2 = [0] * self._model_setting.L
        Xt_to_t1 = [0] * (self._model_setting.L + 1)
        X_layer_split1 = [0] * self._model_setting.L
        X_layer = [0] * self._model_setting.L
        Piprior = [0] * self._model_setting.L

        for l in range(self._model_setting.L):
            if l == 0:
                A_VK[l] = np.zeros((self._model_setting.V, self._model_setting.K[l]))
            else:
                A_VK[l] = np.zeros((self._model_setting.K[l - 1], self._model_setting.K[l]))

            A_KT[l] = np.zeros((self._model_setting.K[l], self._model_setting.T))
            L_dotkt[l] = np.zeros((self._model_setting.K[l], self._model_setting.T + 1))
            L_kdott[l] = np.zeros((self._model_setting.K[l], self._model_setting.T + 1))
            X_layer[l] = np.zeros((self._model_setting.K[l], self._model_setting.T, 2))

        for iter in range(iter_all):

            start_time = time.time()
            for l in range(self._model_setting.L):

                L_KK[l] = np.zeros((self._model_setting.K[l], self._model_setting.K[l]))

                if l == 0:
                    A_KT[l], A_VK[l] = self._sampler.multi_aug(data, self.global_params.Phi[l], self.local_params.Theta[l])
                else:
                    A_KT[l], A_VK[l] = self._sampler.multi_aug(Xt_to_t1[l], self.global_params.Phi[l], self.local_params.Theta[l])

                if l == self._model_setting.L - 1:
                    for t in range(self._model_setting.T - 1, 0, -1):  # T-1 : 1
                        tmp1 = A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2]
                        L_kdott[l][:, t:t + 1] = self._sampler.crt(tmp1, self._hyper_params.tao0 * np.dot(self.global_params.Pi[l], self.local_params.Theta[l][:, t - 1:t]))
                        L_dotkt[l][:, t:t + 1], tmp = self._sampler.multi_aug(L_kdott[l][:, t:t + 1], self.global_params.Pi[l], self.local_params.Theta[l][:, t - 1:t])
                        L_KK[l] = L_KK[l] + tmp

                else:
                    prob1[l] = self._hyper_params.tao0 * np.dot(self.global_params.Pi[l], self.local_params.Theta[l])
                    prob2[l] = self._hyper_params.tao0 * np.dot(self.global_params.Phi[l + 1], self.local_params.Theta[l + 1])
                    X_layer[l] = np.zeros((self._model_setting.K[l], self._model_setting.T, 2))
                    Xt_to_t1[l + 1] = np.zeros((self._model_setting.K[l], self._model_setting.T))
                    X_layer_split1[l] = np.zeros((self._model_setting.K[l], self._model_setting.T))

                    for t in range(self._model_setting.T - 1, 0, -1):
                        L_kdott[l][:, t:t + 1] = self._sampler.crt(A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2], self._hyper_params.tao0 * (np.dot(self.global_params.Phi[l + 1], self.local_params.Theta[l + 1][:, t:t + 1]) + np.dot(self.global_params.Pi[l], self.local_params.Theta[l][:, t - 1:t])))

                        # split layer2 count
                        tmp_input = np.array(L_kdott[l][:, t:t + 1], dtype=np.float64, order='C')
                        tmp1 = prob1[l][:, t - 1:t]
                        tmp2 = prob2[l][:, t:t + 1]
                        [tmp, X_layer[l][:, t, :]] = self._sampler.multi_aug(tmp_input, np.concatenate((tmp1, tmp2), axis=1), np.ones((2, 1)))
                        X_layer_split1[l][:, t] = np.reshape(X_layer[l][:, t, 0], -1)  # ????
                        Xt_to_t1[l + 1][:, t] = np.reshape(X_layer[l][:, t, 1], -1)  # ?????
                        # sample split1
                        tmp_input = np.array(X_layer_split1[l][:, t:t + 1], dtype=np.float64, order='C')
                        L_dotkt[l][:, t:t + 1], tmp = self._sampler.multi_aug(tmp_input, self.global_params.Pi[l], np.array(self.local_params.Theta[l][:, t - 1:t], order='C'))
                        L_KK[l] = L_KK[l] + tmp

                    L_kdott[l][:, 0:1] = self._sampler.crt(A_KT[l][:, 0:1] + L_dotkt[l][:, 1:2], self._hyper_params.tao0 * np.dot(self.global_params.Phi[l + 1], self.local_params.Theta[l + 1][:, 0:1]))
                    Xt_to_t1[l + 1][:, 0:1] = L_kdott[l][:, 0:1]


                if is_train:
                    # Sample Phi
                    self.global_params.Phi[l] = self.Sample_Pi(A_VK[l], self._hyper_params.eta0)

                    # Sample Pi
                    Piprior[l] = np.dot(self.global_params.V[l], np.transpose(self.global_params.V[l]))
                    Piprior[l][np.arange(Piprior[l].shape[0]), np.arange(Piprior[l].shape[1])] = 0
                    Piprior[l] = Piprior[l] + np.diag(np.reshape(self.global_params.Xi[l] * self.global_params.V[l], [self.global_params.V[l].shape[0]]))
                    self.global_params.Pi[l] = self.Sample_Pi(L_KK[l], Piprior[l])

            # Calculate Zeta
            if (self._model_setting.Stationary == 1):
                for l in range(self._model_setting.L):
                    if (l == 0):
                        for t in range(self._model_setting.T - 1, -1, -1):
                            self.local_params.Zeta[l][t] = np.log(1 + self.local_params.Zeta[l][t + 1] + self.local_params.delta[0][t])

                    else:
                        for t in range(self._model_setting.T - 1, -1, -1):
                            self.local_params.Zeta[l][t] = np.log(1 + self.local_params.Zeta[l][t + 1] + self.local_params.Zeta[l - 1][t])

            # Sample Theta  checked
            for l in range(self._model_setting.L - 1, -1, -1):  # L-1 : 0
                if l == self._model_setting.L - 1:
                    for t in range(self._model_setting.T):
                        if t == 0:
                            shape = A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2] + self._hyper_params.tao0 * self.global_params.V[l]
                        else:
                            shape = A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2] + self._hyper_params.tao0 * np.dot(self.global_params.Pi[l], self.local_params.Theta[l][:, t - 1:t])
                        if (l == 0):
                            scale = self.local_params.delta[0][t] + self._hyper_params.tao0 + self._hyper_params.tao0 * self.local_params.Zeta[l][t + 1]
                        else:
                            scale = self.local_params.Zeta[l - 1][t] + self._hyper_params.tao0 + self._hyper_params.tao0 * self.local_params.Zeta[l][t + 1]

                        self.local_params.Theta[l][:, t:t + 1] = self._sampler.gamma(shape) / scale
                else:
                    for t in range(self._model_setting.T):
                        if t == 0:
                            shape = A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2] + self._hyper_params.tao0 * np.dot(self.global_params.Phi[l + 1], self.local_params.Theta[l + 1][:, t:t + 1])
                        else:
                            shape = A_KT[l][:, t:t + 1] + L_dotkt[l][:, t + 1:t + 2] + self._hyper_params.tao0 * (np.dot(self.global_params.Phi[l + 1], self.local_params.Theta[l + 1][:, t:t + 1]) + np.dot(self.global_params.Pi[l], self.local_params.Theta[l][:, t - 1:t]))
                        if (l == 0):
                            scale = self.local_params.delta[0][t] + self._hyper_params.tao0 + self._hyper_params.tao0 * self.local_params.Zeta[0][t + 1]
                        else:
                            scale = self.local_params.Zeta[l - 1][t] + self._hyper_params.tao0 + self._hyper_params.tao0 * self.local_params.Zeta[l][t + 1]

                        self.local_params.Theta[l][:, t:t + 1] = self._sampler.gamma(shape) / scale

            if is_train:
                # Sample Beta check
                for l in range(self._model_setting.L):
                    shape = self._hyper_params.epilson0 + self._hyper_params.gamma0
                    scale = self._hyper_params.epilson0 + np.sum(self.global_params.V[l])
                    self.global_params.beta[l] = self._sampler.gamma(shape) / scale

                    # Sample q checked
                    a = np.sum(L_dotkt[l], axis=1, keepdims=1)
                    a[a == 0] = 1e-10
                    b = self.global_params.V[l] * (self.global_params.Xi[l] + np.repeat(np.sum(self.global_params.V[l]), self._model_setting.K[l], axis=0).reshape([self._model_setting.K[l], 1]) - self.global_params.V[l])
                    b[b == 0] = 1e-10  #
                    self.global_params.q[l] = np.maximum(self._sampler.beta(b, a), realmin)
                    # Sample h checked
                    for k1 in range(self._model_setting.K[l]):
                        for k2 in range(self._model_setting.K[l]):
                            self.global_params.h[l][k1:k1 + 1, k2:k2 + 1] = self._sampler.crt(L_KK[l][k1:k1 + 1, k2:k2 + 1], Piprior[l][k1:k1 + 1, k2:k2 + 1])
                    # Sample Xi checked
                    shape = self._hyper_params.gamma0 / self._model_setting.K[l] + np.trace(self.global_params.h[l])
                    scale = self.global_params.beta[l] - np.dot(np.transpose(self.global_params.V[l]), log_max(self.global_params.q[l]))
                    self.global_params.Xi[l] = self._sampler.gamma(shape) / scale

                # Sample V # check
                for k in range(self._model_setting.K[self._model_setting.L - 1]):
                    L_kdott[self._model_setting.L - 1][k:k + 1, 0:1] = self._sampler.crt(A_KT[self._model_setting.L - 1][k:k + 1, 0:1] + L_dotkt[self._model_setting.L - 1][k:k + 1, 1:2],
                                                                                       np.reshape(self._hyper_params.tao0 * self.global_params.V[self._model_setting.L - 1][k], (1, 1)))
                    self.global_params.n[self._model_setting.L - 1][k] = np.sum(self.global_params.h[self._model_setting.L - 1][k, :] + np.transpose(self.global_params.h[self._model_setting.L - 1][:, k])) \
                                                                         - self.global_params.h[self._model_setting.L - 1][k, k] + L_kdott[self._model_setting.L - 1][k:k + 1, 0:1]
                    self.global_params.rou[self._model_setting.L - 1][k] = - log_max(self.global_params.q[self._model_setting.L - 1][k]) * (self.global_params.Xi[self._model_setting.L - 1] + np.sum(self.global_params.V[self._model_setting.L - 1]) - self.global_params.V[self._model_setting.L - 1][k]) \
                                                                           - np.dot(np.transpose(log_max(self.global_params.q[self._model_setting.L - 1])), self.global_params.V[self._model_setting.L - 1]) \
                                                                           + log_max(self.global_params.q[self._model_setting.L - 1][k]) * self.global_params.V[self._model_setting.L - 1][k] + self.local_params.Zeta[self._model_setting.L - 1][0]
                shape_top = self._hyper_params.gamma0 / self._model_setting.K[self._model_setting.L - 1] + self.global_params.n[self._model_setting.L - 1]
                scale_top = self.global_params.beta[self._model_setting.L - 1] + self.global_params.rou[self._model_setting.L - 1]
                self.global_params.V[self._model_setting.L - 1] = self._sampler.gamma(shape_top) / scale_top

                # Sample V[1] V[L] check
                if (self._model_setting.L > 1):
                    for l in range(self._model_setting.L - 1):
                        for k in range(self._model_setting.K[l]):
                            self.global_params.n[l][k] = np.sum(self.global_params.h[l][k, :] + np.transpose(self.global_params.h[l][:, k])) - self.global_params.h[l][k, k]
                            self.global_params.rou[l][k] = - log_max(self.global_params.q[l][k]) * (self.global_params.Xi[l] + np.sum(self.global_params.V[l]) - self.global_params.V[l][k]) \
                                                     - np.dot(log_max(np.transpose(self.global_params.q[l])), self.global_params.V[l]) + log_max(self.global_params.q[l][k] * self.global_params.V[l][k])

                        shape = self._hyper_params.gamma0 / self._model_setting.K[l] + self.global_params.n[l]
                        scale = self.global_params.beta[l] + self.global_params.rou[l]
                        self.global_params.V[l] = self._sampler.gamma(shape) / scale

            # Likelihood
            deltatmp = np.zeros([self._model_setting.K[0], self._model_setting.T])
            deltatmp[0:self._model_setting.K[0] - 1, :] = np.transpose(self.local_params.delta[0])
            Theta_hat = deltatmp * self.local_params.Theta[0]
            lambd = np.dot(self.global_params.Phi[0], self.local_params.Theta[0])

            like = np.sum(data * np.log(lambd) - lambd) / self._model_setting.V / self._model_setting.T

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds. Likelihood:{like:8.3f}')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params

    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/DPGDS.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')

    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])

        self.Para = self._hyper_params.Para

    def Sample_Pi(self, WSZS, Eta):
        Phi = self._sampler.gamma(WSZS + Eta)
        tmp = np.sum(Phi, axis=0)
        temp_dex = np.where(tmp > 0)
        temp_dex_no = np.where(tmp <= 0)
        Phi[:, temp_dex] = Phi[:, temp_dex] / (tmp[temp_dex] + realmin)
        Phi[:, temp_dex_no] = 0
        return Phi



